#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


from xpra.log import Logger


def dbus_to_native(value):
    # log("dbus_to_native(%s) type=%s", value, type(value))
    if value is None:
        return None
    if isinstance(value, int):
        return int(value)
    if isinstance(value, dict):
        d = {}
        for k, v in value.items():
            d[dbus_to_native(k)] = dbus_to_native(v)
        return d
    if isinstance(value, str):
        return str(value)
    if isinstance(value, float):
        return float(value)
    if isinstance(value, list):
        return [dbus_to_native(x) for x in value]
    from dbus import Struct
    if isinstance(value, Struct):
        return [dbus_to_native(value[i]) for i in range(len(value))]
    return value


def native_to_dbus(value, signature=None):
    try:
        from dbus import types
    except ImportError as e:
        raise RuntimeError(f"the dbus bindings are missing: {e}")
    if value is None:
        return None
    if isinstance(value, int):
        return types.Int64(value)
    if isinstance(value, str):
        return types.String(value)
    if isinstance(value, float):
        return types.Double(value)
    if isinstance(value, (tuple, list, bytearray)):
        if not value:
            return types.Array(signature="s")
        keytypes = set(type(x) for x in value)
        if not signature and len(keytypes) == 1:
            # just one type of key:
            keytype = tuple(keytypes)[0]
            if keytype is int:
                signature = "i"
            elif keytype is bool:
                signature = "b"
            elif keytype is float:
                signature = "d"
        if signature:
            value = [native_to_dbus(v) for v in value]
        else:
            signature = "s"
            # use strings as keys
            value = [native_to_dbus(str(v)) for v in value]
        return types.Array(value, signature=signature)
    if isinstance(value, dict):
        if not value:
            return types.Dictionary({}, signature=signature or "sv")
        if signature is None:
            keytypes = set(type(x) for x in value.keys())
            sig = None
            if len(keytypes) == 1:
                # just one type of key:
                keytype = tuple(keytypes)[0]
                if keytype is int:
                    sig = "i"
                elif keytype is bool:
                    sig = "b"
                elif keytype is float:
                    sig = "d"
            if sig:
                value = {k: native_to_dbus(v) for k, v in value.items()}
            else:
                sig = "s"
                # use strings as keys
                value = {str(k): native_to_dbus(v) for k, v in value.items()}
            signature = f"{sig}v"
        return types.Dictionary(value, signature=signature)
    return types.String(value)


class DBusHelper:

    def __init__(self):
        from xpra.dbus.common import init_session_bus
        self.bus = init_session_bus()

    def get_session_bus(self):
        return self.bus

    def dbus_to_native(self, *args):
        return dbus_to_native(*args)

    def call_function(self, bus_name, path, interface, function, args, ok_cb, err_cb):
        log = Logger("dbus")

        def err(msg):
            log("DBusHelper: %s", msg)
            err_cb(msg)
            return None

        from dbus import DBusException
        try:
            # remote_object = self.bus.get_object("com.example.SampleService","/SomeObject")
            obj = self.bus.get_object(bus_name, path)
            log("dbus.get_object(%s, %s)=%s", bus_name, path, obj)
        except DBusException:
            return err("failed to locate object at: %s:%s" % (bus_name, path))
        try:
            fn = obj.get_dbus_method(function, interface)
            log("%s.get_dbus_method(%s, %s)=%s", obj, function, interface, fn)
        except Exception:
            return err("failed to locate remote function '%s' on %s" % (function, obj))
        try:
            log("calling %s(%s)", fn, args)
            keywords = {"dbus_interface": interface,
                        "reply_handler": ok_cb,
                        "error_handler": err_cb}
            fn.call_async(*args, **keywords)
        except Exception as e:
            return err("error invoking %s on %s: %s" % (function, obj, e))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.scripts.server import is_splash_enabled


class TestMain(unittest.TestCase):

    def test_splash_enabled(self):
        assert is_splash_enabled("foo", True, True, ":10") is False, "splash should not be enabled for daemons"
        assert is_splash_enabled("foo", False, False, ":10") is False, "splash should not be enabled for splash=False"
        assert is_splash_enabled("foo", False, True, ":10") is True, "splash should be enabled for splash=True"


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

from typing import Any
from collections.abc import Callable

from xpra.os_util import gi_import
from xpra.common import NotificationID, noop
from xpra.platform.paths import get_icon_filename
from xpra.platform.gui import get_native_notifier_classes
from xpra.net.common import PacketType
from xpra.util.objects import typedict, make_instance
from xpra.util.str_fn import repr_ellipsized
from xpra.util.env import envbool
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("notify")

NATIVE_NOTIFIER = envbool("XPRA_NATIVE_NOTIFIER", True)
THREADED_NOTIFICATIONS = envbool("XPRA_THREADED_NOTIFICATIONS", True)


class NotificationClient(StubClientMixin):
    """
    Mixin for clients that handle notifications
    """
    PREFIX = "notification"

    def __init__(self):
        self.client_supports_notifications = False
        self.server_notifications = False
        self.notifications_enabled = False
        self.notifier = None
        self.tray = None
        self.callbacks: dict[int, Callable] = {}
        # override the default handler in client base:
        self.may_notify = self.do_notify

    def init(self, opts) -> None:
        if opts.notifications:
            try:
                from xpra import notifications
                assert notifications
            except ImportError:
                log.warn("Warning: notifications module not found")
            else:
                self.client_supports_notifications = True
                self.notifier = self.make_notifier()
                log("using notifier=%s", self.notifier)
                self.client_supports_notifications = self.notifier is not None

    def cleanup(self) -> None:
        n = self.notifier
        log("NotificationClient.cleanup() notifier=%s", n)
        if n:
            self.notifier = None
            with log.trap_error(f"Error on notifier {n!r} cleanup"):
                n.cleanup()

    def parse_server_capabilities(self, c: typedict) -> bool:
        self.server_notifications = "notifications" in c
        self.notifications_enabled = self.client_supports_notifications
        return True

    def get_caps(self) -> dict[str, Any]:
        enabled = self.client_supports_notifications
        return {
            "notifications": {
                "enabled": enabled,
            },
        }

    def init_authenticated_packet_handlers(self) -> None:
        self.add_packets(f"{NotificationClient.PREFIX}-show", f"{NotificationClient.PREFIX}-close")
        self.add_legacy_alias("notify_show", f"{NotificationClient.PREFIX}-show")
        self.add_legacy_alias("notify_close", f"{NotificationClient.PREFIX}-close")

    def make_notifier(self):
        nc = self.get_notifier_classes()
        log("make_notifier() notifier classes: %s", nc)
        return make_instance(nc, self.notification_closed, self.notification_action)

    def notification_closed(self, nid: int, reason=3, text="") -> None:
        log("notification_closed(%i, %i, %s)", nid, reason, text)
        callback = self.callbacks.pop(nid, None)
        if callback:
            callback("notification-close", nid, reason, text)
        else:
            self.send("notification-close", nid, reason, text)

    def notification_action(self, nid: int, action_id: int) -> None:
        log("notification_action(%i, %s)", nid, action_id)
        callback = self.callbacks.get(nid, None)
        if callback:
            callback("notification-action", nid, action_id)
        else:
            self.send("notification-action", nid, action_id)

    def get_notifier_classes(self) -> list[type]:
        # subclasses will generally add their toolkit specific variants
        # by overriding this method
        # use the native ones first:
        if not NATIVE_NOTIFIER:
            return []
        return get_native_notifier_classes()

    def do_notify(self, nid: int | NotificationID, summary: str, body: str, actions=(),
                  hints=None, expire_timeout=10 * 1000, icon_name: str = "", callback=noop) -> None:
        log("do_notify%s client_supports_notifications=%s, notifier=%s",
            (nid, summary, body, actions, hints, expire_timeout, icon_name),
            self.client_supports_notifications, self.notifier)
        if callback:
            self.callbacks[nid] = callback
        n = self.notifier
        if not self.client_supports_notifications or not n:
            # just log it instead:
            log.info("%s", summary)
            if body:
                for x in body.splitlines():
                    log.info(" %s", x)
            return

        def show_notification() -> None:
            try:
                from xpra.notifications.common import parse_image_path
                icon_filename = get_icon_filename(icon_name)
                icon = parse_image_path(icon_filename)
                n.show_notify("", self.tray, int(nid), "Xpra", int(nid), "",
                              summary, body, actions, hints or {}, expire_timeout, icon)
            except Exception as e:
                log("failed to show notification", exc_info=True)
                log.error("Error: cannot show notification")
                log.error(" '%s'", summary)
                log.estr(e)

        if THREADED_NOTIFICATIONS:
            show_notification()
        else:
            GLib.idle_add(show_notification)

    def _process_notification_show(self, packet: PacketType) -> None:
        if not self.notifications_enabled:
            log("process_notify_show: ignoring packet, notifications are disabled")
            return
        self._ui_event()
        dbus_id = packet[1]
        nid = int(packet[2])
        app_name = str(packet[3])
        replaces_nid = int(packet[4])
        app_icon = packet[5]
        summary = str(packet[6])
        body = str(packet[7])
        expire_timeout = int(packet[8])
        icon = None
        actions, hints = [], {}
        if len(packet) >= 10:
            icon = packet[9]
        if len(packet) >= 12:
            actions, hints = packet[10], packet[11]
        # note: if the server doesn't support notification forwarding,
        # it can still send us the messages (via xpra control or the dbus interface)
        log("_process_notification_show(%s) notifier=%s, server_notifications=%s",
            repr_ellipsized(packet), self.notifier, self.server_notifications)
        log("notification actions=%s, hints=%s", actions, hints)
        assert self.notifier
        # this one of the few places where we actually do care about character encoding:
        tray = self.get_tray_window(app_name, hints)
        log("get_tray_window(%s)=%s", app_name, tray)
        self.notifier.show_notify(dbus_id, tray, nid,
                                  app_name, replaces_nid, app_icon,
                                  summary, body, actions, hints, expire_timeout, icon)

    def _process_notification_close(self, packet: PacketType) -> None:
        if not self.notifications_enabled:
            return
        assert self.notifier
        nid = packet[1]
        log("_process_notification_close(%s)", nid)
        self.notifier.close_notify(nid)

    def get_tray_window(self, _app_name, _hints):
        # overridden in subclass to use the correct window if we can find it
        return self.tray



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Sequence

import Quartz.CoreGraphics as CG

from xpra.os_util import gi_import
from xpra.util.env import envbool
from xpra.util.str_fn import memoryview_to_bytes
from xpra.scripts.config import InitExit
from xpra.scripts.main import check_display
from xpra.exit_codes import ExitCode
from xpra.codecs.image import ImageWrapper
from xpra.server.gtk_server import GTKServerBase
from xpra.server.shadow.gtk_shadow_server_base import GTKShadowServerBase
from xpra.platform.darwin.keyboard_config import KeyboardConfig
from xpra.platform.darwin.gui import get_CG_imagewrapper, take_screenshot
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("shadow", "osx")

USE_TIMER = envbool("XPRA_OSX_SHADOW_USE_TIMER", False)
GSTREAMER = envbool("XPRA_SHADOW_GSTREAMER", True)

GSTREAMER_CAPTURE_ELEMENTS: Sequence[str] = ("avfvideosrc", )


def check_gstreamer() -> bool:
    if not GSTREAMER:
        return False
    from xpra.gstreamer.common import has_plugins, import_gst
    import_gst()
    return has_plugins("avfvideosrc")


SHADOW_OPTIONS = {
    "auto": lambda: True,
    "gstreamer": check_gstreamer,
}

ALPHA: dict[int, str] = {
    CG.kCGImageAlphaNone: "AlphaNone",
    CG.kCGImageAlphaPremultipliedLast: "PremultipliedLast",
    CG.kCGImageAlphaPremultipliedFirst: "PremultipliedFirst",
    CG.kCGImageAlphaLast: "Last",
    CG.kCGImageAlphaFirst: "First",
    CG.kCGImageAlphaNoneSkipLast: "SkipLast",
    CG.kCGImageAlphaNoneSkipFirst: "SkipFirst",
}

BTYPES = tuple((str, bytes, memoryview, bytearray))


# ensure that picture_encode can deal with pixels as NSCFData:
def patch_pixels_to_bytes() -> None:
    from CoreFoundation import CFDataGetBytes, CFDataGetLength

    def pixels_to_bytes(v):
        if isinstance(v, BTYPES):
            return memoryview_to_bytes(v)
        size = CFDataGetLength(v)
        return CFDataGetBytes(v, (0, size), None)

    from xpra.codecs import rgb_transform
    rgb_transform.pixels_to_bytes = pixels_to_bytes


class OSXRootCapture:

    def __repr__(self):
        return "OSXRootCapture"

    def get_type(self) -> str:
        return "CoreGraphics"

    def refresh(self) -> bool:
        return True

    def clean(self) -> None:
        """ nothing specific to do here on MacOS """

    def get_image(self, x, y, width, height) -> ImageWrapper:
        rect = (x, y, width, height)
        return get_CG_imagewrapper(rect)

    def get_info(self) -> dict[str, Any]:
        return {
            "type": "CoreGraphics",
        }

    def take_screenshot(self) -> tuple[int, int, str, int, bytes]:
        log("grabbing screenshot")
        return take_screenshot()


class ShadowServer(GTKShadowServerBase):

    def __init__(self, display: str, attrs: dict[str, str]):
        super().__init__(attrs)
        # sanity check:
        check_display()
        image = CG.CGWindowListCreateImage(CG.CGRectInfinite,
                                           CG.kCGWindowListOptionOnScreenOnly,
                                           CG.kCGNullWindowID,
                                           CG.kCGWindowImageDefault)
        if image is None:
            log("cannot grab test screenshot - maybe you need to run this command whilst logged in via the UI")
            raise InitExit(ExitCode.FAILURE,
                           "cannot grab pixels from the screen, make sure this command is launched from a GUI session")
        patch_pixels_to_bytes()
        self.refresh_count = 0
        self.refresh_rectangle_count = 0
        self.refresh_registered = False
        super().__init__(attrs)

    def init(self, opts) -> None:
        super().init(opts)
        # printing fails silently on OSX
        self.printing = False

    def get_keyboard_config(self, _props=None) -> KeyboardConfig:
        return KeyboardConfig()

    def make_tray_widget(self):
        from xpra.client.gtk3.statusicon_tray import GTKStatusIconTray
        return GTKStatusIconTray(self, 0, self.tray, "Xpra Shadow Server",
                                 click_cb=self.tray_click_callback, mouseover_cb=None, exit_cb=self.tray_exit_callback)

    def setup_capture(self) -> OSXRootCapture:
        return OSXRootCapture()

    def screen_refresh_callback(self, count, rects, info) -> None:
        log("screen_refresh_callback%s mapped=%s", (count, rects, info), self.mapped)
        self.refresh_count += 1
        rlist = []
        for r in rects:
            if not isinstance(r, CG.CGRect):
                log.error("Error: invalid rectangle in refresh list: %s", r)
                continue
            self.refresh_rectangle_count += 1
            rlist.append((int(r.origin.x), int(r.origin.y), int(r.size.width), int(r.size.height)))
        # return quickly, and process the list copy via idle add:
        GLib.idle_add(self.do_screen_refresh, rlist)

    def do_screen_refresh(self, rlist: list) -> None:
        # TODO: improve damage method to handle lists directly:
        from xpra.util.rectangle import rectangle
        model_rects = {}
        for model in self._id_to_window.values():
            model_rects[model] = rectangle(*model.geometry)
        for x, y, w, h in rlist:
            for model, rect in model_rects.items():
                mrect = rect.intersection(x, y, w, h)
                # log("screen refresh intersection of %s and %24s: %s", model, (x, y, w, h), mrect)
                if mrect:
                    rx = mrect.x - rect.x
                    ry = mrect.y - rect.y
                    self.refresh_window_area(model, rx, ry, mrect.width, mrect.height, {"damage": True})

    def start_refresh(self, wid: int) -> None:
        # don't use the timer, get damage notifications:
        if wid not in self.mapped:
            self.mapped.append(wid)
        if self.refresh_registered:
            return
        if not USE_TIMER:
            err = CG.CGRegisterScreenRefreshCallback(self.screen_refresh_callback, None)
            log("CGRegisterScreenRefreshCallback(%s)=%s", self.screen_refresh_callback, err)
            if err == 0:
                self.refresh_registered = True
                return
            log.warn("Warning: CGRegisterScreenRefreshCallback failed with error %i", err)
            log.warn(" using fallback timer method")
        super().start_refresh(wid)

    def stop_refresh(self, wid: int) -> None:
        log("stop_refresh(%i) mapped=%s, timer=%s", wid, self.mapped, self.refresh_timer)
        # may stop the timer fallback:
        super().stop_refresh(wid)
        if self.refresh_registered and not self.mapped:
            try:
                err = CG.CGUnregisterScreenRefreshCallback(self.screen_refresh_callback, None)
                log("CGUnregisterScreenRefreshCallback(%s)=%s", self.screen_refresh_callback, err)
                if err:
                    log.warn(" unregistering the existing one returned %s", {0: "OK"}.get(err, err))
            except ValueError as e:
                log.warn("Error unregistering screen refresh callback:")
                log.warn(" %s", e)
            self.refresh_registered = False

    def do_process_mouse_common(self, proto, device_id: int, wid: int, pointer, props) -> bool:
        if proto not in self._server_sources:
            return False
        assert wid in self._id_to_window
        CG.CGWarpMouseCursorPosition(pointer[:2])
        return True

    def fake_key(self, keycode: int, press: bool) -> None:
        e = CG.CGEventCreateKeyboardEvent(None, keycode, press)
        log("fake_key(%s, %s)", keycode, press)
        # CGEventSetFlags(keyPress, modifierFlags)
        # modifierFlags: kCGEventFlagMaskShift, ...
        CG.CGEventPost(CG.kCGSessionEventTap, e)
        # this causes crashes, don't do it!
        # CG.CFRelease(e)

    def do_process_button_action(self, proto, device_id: int, wid: int, button: int, pressed: bool, pointer, props):
        if "modifiers" in props:
            self._update_modifiers(proto, wid, props.get("modifiers"))
        pointer = self.process_mouse_common(proto, device_id, wid, pointer)
        if pointer:
            self.button_action(device_id, wid, pointer, button, pressed, props)

    def button_action(self, device_id: int, wid: int, pointer, button: int, pressed: bool, props):
        if button <= 3:
            # we should be using CGEventCreateMouseEvent
            # instead we clear previous clicks when a "higher" button is pressed... oh well
            event = [pointer[:2], 1, button]
            for i in range(button):
                event.append(i == (button - 1) and pressed)
            r = CG.CGPostMouseEvent(*event)
            log("CG.CGPostMouseEvent%s=%s", event, r)
            return
        if not pressed:
            # we don't simulate press/unpress
            # so just ignore unpressed events
            return
        wheel = (button - 2) // 2
        direction = 1 - (((button - 2) % 2) * 2)
        event = [wheel]
        for i in range(wheel):
            if i != (wheel - 1):
                event.append(0)
            else:
                event.append(direction)
        r = CG.CGPostScrollWheelEvent(*event)
        log("CG.CGPostScrollWheelEvent%s=%s", event, r)

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = GTKServerBase.make_hello(self, source)
        capabilities["shadow"] = True
        capabilities["server_type"] = "Python/MacOS-Shadow"
        return capabilities

    def get_info(self, proto, *_args) -> dict[str, Any]:
        info = GTKServerBase.get_info(self, proto)
        info.setdefault("features", {})["shadow"] = True
        info.setdefault("server", {})["type"] = "Python/gtk2/osx-shadow"
        info.setdefault("damage", {}).update({
            "use-timer": USE_TIMER,
            "notifications": self.refresh_registered,
            "count": self.refresh_count,
            "rectangles": self.refresh_rectangle_count,
        })
        return info


def main():
    import sys
    from xpra.platform import program_context
    with program_context("MacOS Shadow Capture"):
        log.enable_debug()
        c = OSXRootCapture()
        x, y, w, h = list(int(sys.argv[x]) for x in range(1, 5))
        img = c.get_image(x, y, w, h)
        from PIL import Image
        i = Image.frombuffer("RGBA", (w, h), img.get_pixels(), "raw", "BGRA", img.get_rowstride())
        import time
        t = time.time()
        tstr = time.strftime("%H-%M-%S", time.localtime(t))
        filename = "./Capture-{}-{}.png".format((x, y, w, h), tstr)
        i.save(filename, "png")
        print("saved to {}".format(filename))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import logging
from typing import Any
from collections.abc import Sequence

from xpra.util.str_fn import csv, print_nested_dict
from xpra.util.env import envint, envbool, numpy_import_context
from xpra.log import Logger, CaptureHandler, consume_verbose_argv
from xpra.opengl.drivers import GL_MATCH_LIST, WHITELIST, GREYLIST, BLOCKLIST, OpenGLFatalError

log = Logger("opengl")

required_extensions: Sequence[str] = ("GL_ARB_texture_rectangle", "GL_ARB_vertex_program")

GL_ALPHA_SUPPORTED: bool = envbool("XPRA_ALPHA", True)
DOUBLE_BUFFERED: bool = envbool("XPRA_OPENGL_DOUBLE_BUFFERED", True)

MIN_SIZE = envint("XPRA_OPENGL_MIN_SIZE", 4 * 1024)

CRASH: bool = envbool("XPRA_OPENGL_FORCE_CRASH", False)
TIMEOUT: int = envint("XPRA_OPENGL_FORCE_TIMEOUT", 0)


# by default, we raise an ImportError as soon as we find something missing:
def raise_error(msg) -> None:
    raise ImportError(msg)


def raise_fatal_error(msg) -> None:
    raise OpenGLFatalError(msg)


gl_check_error = raise_error
gl_fatal_error = raise_fatal_error


def parse_pyopengl_version(vstr: str) -> tuple[int, ...]:
    def numv(s):
        try:
            return int(s)
        except ValueError:
            return 0

    return tuple(numv(x) for x in vstr.split("."))


_version_warning_shown = False


def get_max_texture_size() -> int:
    # pylint: disable=import-outside-toplevel
    from OpenGL.GL import glGetInteger, GL_MAX_TEXTURE_SIZE
    texture_size = glGetInteger(GL_MAX_TEXTURE_SIZE)
    log("GL_MAX_TEXTURE_SIZE=%s", texture_size)
    # this one may be missing?
    rect_texture_size = texture_size
    try:
        from OpenGL.GL import GL_MAX_RECTANGLE_TEXTURE_SIZE
        rect_texture_size = glGetInteger(GL_MAX_RECTANGLE_TEXTURE_SIZE)
    except ImportError as e:
        log("OpenGL: %s", e)
        log("using GL_MAX_TEXTURE_SIZE=%s as default", texture_size)
    except Exception as e:
        log("failed to query GL_MAX_RECTANGLE_TEXTURE_SIZE: %s", e)
    else:
        log("Texture size GL_MAX_RECTANGLE_TEXTURE_SIZE=%s", rect_texture_size)
    return int(min(rect_texture_size, texture_size))


def get_array_handlers() -> Sequence[str]:
    from OpenGL.arrays.arraydatatype import ArrayDatatype
    try:
        return tuple(getattr(atype, "__name__", str(atype)) for atype in set(ArrayDatatype.getRegistry().values()))
    except Exception:
        pass
    return ()


def get_max_viewport_dims() -> tuple[int, int]:
    from OpenGL.GL import glGetIntegerv, GL_MAX_VIEWPORT_DIMS
    v = glGetIntegerv(GL_MAX_VIEWPORT_DIMS)
    max_viewport_dims = int(v[0]), int(v[1])
    log("GL_MAX_VIEWPORT_DIMS=%s", max_viewport_dims)
    return max_viewport_dims


def get_extensions() -> list[str]:
    extensions: list[str] = _get_gl_enums("extensions", "GL_NUM_EXTENSIONS", "GL_EXTENSIONS")
    if not extensions:
        # try legacy mode:
        from OpenGL.error import GLError
        try:
            from OpenGL.GL import glGetString, GL_EXTENSIONS
            extensions = glGetString(GL_EXTENSIONS).decode().split(" ")
        except GLError as e:
            log(f"error querying extensions using glGetString(GL_EXTENSIONS): {e}")
    log("OpenGL extensions found: %s", csv(extensions))
    return extensions


def _get_gl_enums(name: str, num_const: str, values_const: str) -> list[str]:
    values: list[str] = []
    from OpenGL.error import GLError
    from OpenGL import GL
    from OpenGL.GL import glGetStringi, glGetIntegerv
    num_enum = getattr(GL, num_const, None)
    values_enum = getattr(GL, values_const, None)
    if num_enum is None or values_enum is None:
        log(f"cannot query {name} using {num_const} / {values_const}: constants not found!")
        return values
    try:
        num = glGetIntegerv(num_enum)
        for i in range(num):
            values.append(glGetStringi(values_enum, i).decode("latin1"))
        log(f"OpenGL {name} found: " + csv(values))
    except GLError as e:
        log(f"error querying {name} using {num_const} / {values_const}: {e}")
    return values


def get_shader_binary_formats() -> list[str]:
    return _get_gl_enums("shader binary formats", "GL_NUM_SHADER_BINARY_FORMATS", "GL_SHADER_BINARY_FORMATS")


def get_program_binary_formats() -> list[str]:
    return _get_gl_enums("program binary formats", "GL_NUM_PROGRAM_BINARY_FORMATS", "GL_PROGRAM_BINARY_FORMATS")


def fixstring(v) -> str:
    try:
        return str(v).strip()
    except Exception:
        return str(v)


def get_vendor_info() -> dict[str, str]:
    from OpenGL.GL import GL_RENDERER, GL_VENDOR, GL_SHADING_LANGUAGE_VERSION
    from OpenGL.GL import glGetString
    info: dict[str, str] = {}
    for d, s, fatal in (
            ("vendor", GL_VENDOR, True),
            ("renderer", GL_RENDERER, True),
            ("shading-language-version", GL_SHADING_LANGUAGE_VERSION, False)
    ):
        try:
            v = glGetString(s)
            v = fixstring(v.decode())
            log("%s: %s", d, v)
            info[d] = v
        except Exception:
            if fatal:
                raise RuntimeError(f"OpenGL property {d!r} is missing")
            log(f"OpenGL property {d!r} is missing")
    return info


def get_GLU_info() -> dict[str, str]:
    props: dict[str, str] = {}
    from OpenGL.GLU import gluGetString, GLU_VERSION, GLU_EXTENSIONS
    # maybe we can continue without?
    if not bool(gluGetString):
        return props
    for d, s in {
        "GLU": GLU_VERSION,
        "GLU.extensions": GLU_EXTENSIONS,
    }.items():
        v = gluGetString(s)
        v = fixstring(v.decode())
        log("%s: %s", d, v)
        if v:
            props[d] = v
    return props


def get_context_info() -> dict[str, Any]:
    with numpy_import_context("OpenGL: context info", True):
        from OpenGL.GL import glGetIntegerv
        from OpenGL.GL import (
            GL_CONTEXT_PROFILE_MASK, GL_CONTEXT_CORE_PROFILE_BIT,
            GL_CONTEXT_FLAGS,
            GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT,
            GL_CONTEXT_FLAG_ROBUST_ACCESS_BIT,
            GL_CONTEXT_FLAG_DEBUG_BIT,
            GL_CONTEXT_FLAG_NO_ERROR_BIT,
        )
        flags = glGetIntegerv(GL_CONTEXT_FLAGS)
        return {
            "core-profile": bool(glGetIntegerv(GL_CONTEXT_PROFILE_MASK) & GL_CONTEXT_CORE_PROFILE_BIT),
            "flags": tuple(k for k, v in {
                "forward-compatible": GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT,
                "debug": GL_CONTEXT_FLAG_DEBUG_BIT,
                "robust-access": GL_CONTEXT_FLAG_ROBUST_ACCESS_BIT,
                "no-error": GL_CONTEXT_FLAG_NO_ERROR_BIT,
            }.items() if flags & v),
        }


def check_available(*functions) -> str:
    missing = []
    available = []
    for x in functions:
        try:
            name = x.__name__
        except AttributeError:
            name = str(x)
        if not bool(x):
            missing.append(name)
        else:
            available.append(name)
    if missing:
        log("some functions are missing: %s", csv(missing))
        return "missing functions: " + csv(missing)
    else:
        log("All the required OpenGL functions are available: %s " % csv(available))
    return ""


def check_base_functions() -> str:
    # check for specific functions we need:
    from OpenGL.GL import (
        glActiveTexture, glTexSubImage2D, glTexCoord2i,
        glViewport, glMatrixMode, glLoadIdentity, glOrtho,
        glEnableClientState, glGenTextures, glDisable,
        glBindTexture, glPixelStorei, glEnable, glFlush,
        glTexParameteri, glTexEnvi, glHint, glBlendFunc, glLineStipple,
        glTexImage2D,
        glMultiTexCoord2i,
        glVertex2i,
    )
    return check_available(
        glActiveTexture, glTexSubImage2D, glTexCoord2i,
        glViewport, glMatrixMode, glLoadIdentity, glOrtho,
        glEnableClientState, glGenTextures, glDisable,
        glBindTexture, glPixelStorei, glEnable, glFlush,
        glTexParameteri, glTexEnvi, glHint, glBlendFunc, glLineStipple,
        glTexImage2D,
        glMultiTexCoord2i,
        glVertex2i,
    )


def check_framebuffer_functions() -> str:
    # check for framebuffer functions we need:
    from OpenGL.GL.ARB.framebuffer_object import (
        GL_FRAMEBUFFER, GL_DRAW_FRAMEBUFFER, GL_READ_FRAMEBUFFER,
        GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1,
        glGenFramebuffers, glBindFramebuffer, glFramebufferTexture2D,
    )
    return check_available(
        GL_FRAMEBUFFER, GL_DRAW_FRAMEBUFFER, GL_READ_FRAMEBUFFER,
        GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1,
        glGenFramebuffers, glBindFramebuffer, glFramebufferTexture2D,
    )


def check_shader_functions() -> str:
    from OpenGL.GL import (
        glCreateShader, glDeleteShader,
        glShaderSource, glCompileShader, glGetShaderiv, glGetShaderInfoLog,
        glDeleteProgram,
        GL_FRAGMENT_SHADER, GL_COMPILE_STATUS,
    )
    # don't check GL_FALSE, which is zero!
    return check_available(
        glCreateShader, glDeleteShader,
        glShaderSource, glCompileShader, glGetShaderiv, glGetShaderInfoLog,
        glDeleteProgram,
        GL_FRAGMENT_SHADER, GL_COMPILE_STATUS,
    )


def check_texture_functions() -> str:
    from OpenGL.GL.ARB.texture_rectangle import glInitTextureRectangleARB
    return check_available(glInitTextureRectangleARB)


def match_list(props: dict[str, Any], thelist: GL_MATCH_LIST, listname: str) -> tuple[str, str] | None:
    for k, values in thelist.items():
        prop = str(props.get(k))
        if prop and any(True for x in values if prop.find(x) >= 0):
            log("%s '%s' found in %s: %s", k, prop, listname, values)
            return k, prop
        log("%s '%s' not found in %s: %s", k, prop, listname, values)
    return None


def check_lists(props: dict[str, Any], force_enable=False) -> bool:
    blocklisted = match_list(props, BLOCKLIST, "blocklist")
    greylisted = match_list(props, GREYLIST, "greylist")
    whitelisted = match_list(props, WHITELIST, "whitelist")
    if blocklisted:
        if whitelisted:
            log.info("%s '%s' enabled (found in both blocklist and whitelist)", *whitelisted)
        elif force_enable:
            log.info("OpenGL %s '%s' is blocklisted!", *blocklisted)
            log.info(" force enabled by option")
        else:
            log.info("OpenGL %s '%s' is blocklisted!", *blocklisted)
            raise_fatal_error("%s '%s' is blocklisted!" % blocklisted)
    if greylisted and not whitelisted:
        log.info("OpenGL %s '%s' is greylisted,", *greylisted)
        log.info(" you may want to turn off OpenGL if you encounter bugs")
    return bool(whitelisted) or not bool(blocklisted)


def check_PyOpenGL_support(force_enable) -> dict[str, Any]:
    redirected_loggers: dict[str, tuple[Logger, list, bool]] = {}
    try:
        if CRASH:
            import ctypes
            ctypes.string_at(0)
            raise RuntimeError("should have crashed!")
        if TIMEOUT > 0:
            import time
            time.sleep(TIMEOUT)

        # log redirection:
        for name in ("formathandler", "extensions", "acceleratesupport", "arrays", "converters", "plugins"):
            logger = logging.getLogger(f"OpenGL.{name}")
            redirected_loggers[name] = (logger, list(logger.handlers), logger.propagate)
            logger.handlers = [CaptureHandler()]
            logger.propagate = False
        log(f"{redirected_loggers=}")

        with numpy_import_context("OpenGL: check", True):
            return do_check_PyOpenGL_support(force_enable)

    finally:
        def recs(rname) -> list[str]:
            rlog = redirected_loggers.get(rname)
            if not rlog:
                return []
            records = rlog[0].handlers[0].records
            return list(rec.getMessage() for rec in records)

        for msg in recs("acceleratesupport"):
            # strip default message prefix:
            msg = msg.replace("No OpenGL_accelerate module loaded: ", "")
            if msg == "No module named OpenGL_accelerate":
                msg = "missing accelerate module"
            if msg == "OpenGL_accelerate module loaded":
                log.info(msg)
            else:
                log.warn("PyOpenGL warning: %s", msg)

        # format handler messages:
        strip_log_message = "Unable to load registered array format handler "
        missing_handlers = []
        for msg in recs("formathandler"):
            log.info(f"msg formathandler={msg}")
            p = msg.find(strip_log_message)
            if p < 0:
                # unknown message, log it:
                log.info(msg)
                continue
            format_handler = msg[p + len(strip_log_message):]
            p = format_handler.find(":")
            if p > 0:
                format_handler = format_handler[:p]
                missing_handlers.append(format_handler)
        if missing_handlers:
            log.warn("PyOpenGL warning: missing array format handlers: %s", csv(missing_handlers))

        for msg in recs("extensions"):
            # hide extension messages:
            if msg.startswith("GL Extension ") or msg.endswith("available"):
                log(msg)
            else:
                log.info(msg)

        missing_accelerators = []
        strip_ar_head = "Unable to load"
        strip_ar_tail = "from OpenGL_accelerate"
        for msg in recs("arrays") + recs("converters"):
            if msg.startswith(strip_ar_head) and msg.endswith(strip_ar_tail):
                m = msg[len(strip_ar_head):-len(strip_ar_tail)].strip()
                m = m.replace("accelerators", "").replace("accelerator", "").strip()
                missing_accelerators.append(m)
                continue
            if msg.startswith("Using accelerated"):
                log(msg)
            else:
                log.info(msg)
        if missing_accelerators:
            missing_str = csv(missing_accelerators)
            if missing_str == "numpy_formathandler":
                log_fn = log.debug
            else:
                log_fn = log.info
            log_fn(f"OpenGL accelerate missing: {missing_str}")

        for msg in recs("plugins"):
            log(f"plugins msg={msg}")

        for logger, handlers, propagate in redirected_loggers.values():
            logger.handlers = handlers
            logger.propagate = propagate


def do_check_PyOpenGL_support(force_enable) -> dict[str, Any]:
    props: dict[str, Any] = {
        "platform": sys.platform,
    }
    try:
        from OpenGL import platform
        props["backend"] = platform.PLATFORM.__module__.split(".")[-1]
    except (AttributeError, IndexError):
        pass

    def unsafe(warning: str):
        log(f"unsafe: {warning}")
        props["safe"] = False
        warnings = props.get("warning", [])
        warnings.append(warning)
        props["warning"] = warnings

    import OpenGL
    props["pyopengl"] = OpenGL.__version__  # @UndefinedVariable
    from OpenGL.GL import GL_VERSION, glGetString
    gl_version_str = (glGetString(GL_VERSION) or b"").decode("latin1")
    if not gl_version_str and not force_enable:
        raise_fatal_error("OpenGL version is missing - cannot continue")
        return props
    # '4.6.0 NVIDIA 440.59' -> ['4', '6', '0 NVIDIA...']
    log("GL_VERSION=%s", gl_version_str)
    vparts = gl_version_str.split(" ", 1)[0].split(".")
    try:
        gl_major = int(vparts[0])
        gl_minor = int(vparts[1])
    except (IndexError, ValueError) as e:
        msg = "failed to parse gl version '%s': %s" % (gl_version_str, e)
        unsafe(msg)
        log(" assuming this is at least 1.1 to continue")
    else:
        props["opengl"] = gl_major, gl_minor
        min_version = (1, 1)
        if (gl_major, gl_minor) < min_version:
            req_vstr = ".".join([str(x) for x in min_version])
            msg = f"OpenGL output requires version {req_vstr} or greater, not {gl_major}.{gl_minor}"
            unsafe(msg)
        else:
            log(f"found valid OpenGL version: {gl_major}.{gl_minor}")

    from OpenGL import version as opengl_version
    pyopengl_version = opengl_version.__version__
    try:
        import OpenGL_accelerate
        accel_version = OpenGL_accelerate.__version__
        props["accelerate"] = accel_version
        log(f"OpenGL_accelerate version {accel_version}")
    except ImportError:
        log("OpenGL_accelerate not found")
        accel_version = ""

    if accel_version and pyopengl_version != accel_version:
        global _version_warning_shown
        if not _version_warning_shown:
            log.warn("Warning: version mismatch between PyOpenGL and PyOpenGL-accelerate")
            log.warn(" %s vs %s", pyopengl_version, accel_version)
            log.warn(" this may cause crashes")
            _version_warning_shown = True
            gl_check_error(f"PyOpenGL vs accelerate version mismatch: {pyopengl_version} vs {accel_version}")
    vernum = parse_pyopengl_version(pyopengl_version)
    # we now require PyOpenGL 3.1.4 or later
    # 3.1.4 was released in 2019
    if vernum < (3, 1, 4):
        msg = f"PyOpenGL version {pyopengl_version} is too old and buggy"
        unsafe(msg)
        if not force_enable:
            raise_fatal_error(msg)
            return props
    props["zerocopy"] = bool(accel_version and pyopengl_version == accel_version)

    props.update(get_vendor_info())
    props.update(get_GLU_info())
    props.update({
        "extensions": get_extensions(),
        "array-handlers": get_array_handlers(),
        "texture-size-limit": get_max_texture_size(),
        "max-viewport-dims": get_max_viewport_dims(),
    })
    # props["shader-binary-formats"] = get_shader_binary_formats()
    # props["program-binary-formats"] = get_program_binary_formats()

    for check_fn in (
            check_base_functions,
            check_framebuffer_functions,
            check_texture_functions,
            check_shader_functions,
    ):
        msg = check_fn()
        if msg:
            unsafe(msg)

    safe = check_lists(props, force_enable)
    if "safe" not in props:
        props["safe"] = safe
    if safe and match_list(props, GREYLIST, "greylist"):
        props["enable"] = False
        props["message"] = "driver found in greylist"
    try:
        props.update(get_context_info())
    except Exception as e:
        unsafe(f"error querying context flags: {e}")
    return props


def main() -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.platform.gui import init as gui_init
    from xpra.log import enable_color
    with program_context("OpenGL-Check"):
        gui_init()
        enable_color()
        consume_verbose_argv(sys.argv, "opengl")
        from xpra.gtk.util import init_display_source
        init_display_source()
        force_enable = "-f" in sys.argv or "--force" in sys.argv
        from xpra.platform.gl_context import GLContext
        if not GLContext:
            log.warn("No OpenGL context implementation found")
            return 1
        log("testing %s", GLContext)
        gl_context = GLContext()  # pylint: disable=not-callable
        log("GLContext=%s", gl_context)
        # replace ImportError with a log message:
        global gl_check_error, gl_fatal_error
        errors = []

        def log_error(msg):
            log.error("ERROR: %s", msg)
            errors.append(msg)

        gl_check_error = log_error
        gl_fatal_error = log_error
        try:
            props = gl_context.check_support(force_enable)
        except Exception as e:
            props = {}
            log("check_support", exc_info=True)
            errors.append(e)
        log.info("")
        if errors:
            log.info("OpenGL errors:")
            for err in errors:
                log.info("  %s", err)
        if props:
            log.info("")
            log.info("OpenGL properties:")
            print_nested_dict(props)
        return len(errors)


if __name__ == "__main__":
    sys.exit(main())



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Sequence

from xpra.common import noop
from xpra.os_util import POSIX, OSX
from xpra.util.objects import typedict
from xpra.util.env import envint
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.log import Logger

log = Logger("webcam")

MAX_WEBCAM_DEVICES = envint("XPRA_MAX_WEBCAM_DEVICES", 1)


def valid_encodings(args: Sequence[str]) -> list[str]:
    # ensure that the encodings specified can be validated using HEADERS
    try:
        from xpra.codecs.pillow.decoder import HEADERS  # pylint: disable=import-outside-toplevel
    except ImportError:
        return []
    encodings = []
    for x in args:
        if x not in HEADERS.values():
            log.warn("Warning: %s is not supported for webcam forwarding", x)
        else:
            encodings.append(x)
    return encodings


def find_csc_spec(src_format: str, dst_format: str):
    from xpra.codecs.video import getVideoHelper
    specs = getVideoHelper().get_csc_specs(src_format).get(dst_format, ())
    log.error(f"{specs=}")
    for spec in specs:
        if src_format not in spec.input_colorspace:
            continue
        if dst_format not in spec.output_colorspaces:
            continue
        return spec
    raise ValueError("cannot convert %r to %r", src_format, dst_format)


class WebcamMixin(StubSourceMixin):
    """
    Handle webcam forwarding.
    """

    PREFIX = "webcam"

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        if not caps.boolget(WebcamMixin.PREFIX):
            return False
        try:
            from xpra.codecs.pillow.decoder import HEADERS  # pylint: disable=import-outside-toplevel
            assert HEADERS
        except ImportError:
            return False
        return True

    def __init__(self):
        self.webcam_enabled = False
        self.webcam_device = None
        self.webcam_encodings = []

    def init_from(self, _protocol, server) -> None:
        self.webcam_enabled = server.webcam_enabled
        self.webcam_device = server.webcam_device
        self.webcam_encodings = valid_encodings(server.webcam_encodings)
        log("WebcamMixin: enabled=%s, device=%s, encodings=%s",
            self.webcam_enabled, self.webcam_device, self.webcam_encodings)

    def init_state(self) -> None:
        # for each webcam device_id, the actual device used
        self.webcam_forwarding_devices: dict = {}

    def cleanup(self) -> None:
        self.stop_all_virtual_webcams()

    def get_info(self) -> dict[str, Any]:
        return {
            WebcamMixin.PREFIX: {
                "encodings": self.webcam_encodings,
                "active-devices": len(self.webcam_forwarding_devices),
            }
        }

    def get_device_options(self, device_id: int) -> dict[Any, Any]:  # pylint: disable=unused-argument
        if not POSIX or OSX or not self.webcam_enabled:
            return {}
        if self.webcam_device:
            # use the device specified:
            return {
                0: {
                    "device": self.webcam_device,
                },
            }
        from xpra.platform.posix.webcam import get_virtual_video_devices  # pylint: disable=import-outside-toplevel
        return get_virtual_video_devices()

    def send_webcam_ack(self, device, frame: int, *args) -> None:
        self.send_async("webcam-ack", device, frame, *args)

    def send_webcam_stop(self, device, message) -> None:
        self.send_async("webcam-stop", device, message)

    def start_virtual_webcam(self, device_id, w: int, h: int) -> bool:
        log("start_virtual_webcam%s", (device_id, w, h))
        assert w > 0 and h > 0
        webcam = self.webcam_forwarding_devices.get(device_id)
        if webcam:
            log.warn("Warning: virtual webcam device %s already in use,", device_id)
            log.warn(" stopping it first")
            self.stop_virtual_webcam(device_id)

        def fail(msg) -> None:
            log.error("Error: cannot start webcam forwarding")
            log.error(" %s", msg)
            self.send_webcam_stop(device_id, msg)

        if not self.webcam_enabled:
            fail("webcam forwarding is disabled")
            return False
        devices = self.get_device_options(device_id)
        if not devices:
            fail("no virtual devices found")
            return False
        if len(self.webcam_forwarding_devices) > MAX_WEBCAM_DEVICES:
            ndev = len(self.webcam_forwarding_devices)
            fail(f"too many virtual devices are already in use: {ndev}")
            return False
        errs: dict[str, str] = {}
        for vid, device_info in devices.items():
            log("trying device %s: %s", vid, device_info)
            device_str = device_info.get("device")
            try:
                # pylint: disable=import-outside-toplevel
                from xpra.codecs.v4l2.virtual import VirtualWebcam, get_input_colorspaces
                in_cs = get_input_colorspaces()
                p = VirtualWebcam()
                src_format = in_cs[0]
                p.init_context(w, h, w, src_format, device_str)
                self.webcam_forwarding_devices[device_id] = p
                log.info("webcam forwarding using %s", device_str)
                # this tell the client to start sending, and the size to use - which may have changed:
                self.send_webcam_ack(device_id, 0, p.get_width(), p.get_height())
                return True
            except Exception as e:
                log.error("Error: failed to start virtual webcam")
                log.error(" using device %s: %s", vid, device_info, exc_info=True)
                errs[device_str] = str(e)
                del e
        fail("all devices failed")
        if len(errs) > 1:
            log.error(" tried %i devices:", len(errs))
        for device_str, err in errs.items():
            log.error(" %s : %s", device_str, err)
        return False

    def stop_all_virtual_webcams(self) -> None:
        log("stop_all_virtual_webcams() stopping: %s", self.webcam_forwarding_devices)
        for device_id in tuple(self.webcam_forwarding_devices.keys()):
            self.stop_virtual_webcam(device_id)

    def stop_virtual_webcam(self, device_id, message="") -> None:
        webcam = self.webcam_forwarding_devices.pop(device_id, None)
        log("stop_virtual_webcam(%s, %s) webcam=%s", device_id, message, webcam)
        if not webcam:
            log.warn("Warning: cannot stop webcam device %s: no such context!", device_id)
            return
        try:
            webcam.clean()
        except Exception as e:
            log.error("Error stopping virtual webcam device: %s", e)
            log("%s.clean()", exc_info=True)

    def process_webcam_frame(self, device_id, frame_no: int, encoding: str, w: int, h: int, data, options: dict) -> bool:
        webcam = self.webcam_forwarding_devices.get(device_id)
        log("process_webcam_frame: device %s, frame no %i: %s %ix%i, %i bytes, options=%s, webcam=%s",
            device_id, frame_no, encoding, w, h, len(data), options, webcam)
        if not (encoding and w and h and (data or (encoding == "mmap" and options))):
            log.error("Error: webcam frame data is incomplete")
            self.send_webcam_stop(device_id, "incomplete frame data")
            return False
        if not webcam:
            log.error("Error: webcam forwarding is not active, dropping frame")
            self.send_webcam_stop(device_id, "not started")
            return False
        free = noop
        try:
            if encoding == "mmap":
                chunks = options.pop("chunks", ())
                mmap_read_area = getattr(self, "mmap_read_area", None)
                assert mmap_read_area, "no mmap read area!"
                assert self.mmap_supported, "mmap is not supported, yet the client used it!?"
                pixels, free = mmap_read_area.mmap_read(*chunks)
                rgb_pixel_format = options.get("pixel-format", "BGRX")
            else:
                # pylint: disable=import-outside-toplevel
                from xpra.codecs.pillow.decoder import open_only
                if encoding not in self.webcam_encodings:
                    raise ValueError(f"invalid encoding specified: {encoding} (must be one of {self.webcam_encodings})")
                img = open_only(data, (encoding,))
                if img.mode != "RGBA":
                    img = img.convert("RGBA")
                pixels = img.tobytes("raw", "RGBA")
                rgb_pixel_format = "BGRX"

            from xpra.codecs.image import ImageWrapper
            bgrx_image = ImageWrapper(0, 0, w, h, pixels, rgb_pixel_format, 32, w * 4, planes=ImageWrapper.PACKED)
            src_format = webcam.get_src_format()
            if not src_format:
                log("no webcam src format")
                # closed / closing
                return False
            try:
                csc_spec = find_csc_spec(rgb_pixel_format, src_format)
            except ValueError as e:
                message = f"cannot convert {rgb_pixel_format!r} to {src_format!r}: {e}"
                log(f"Warning: {message}")
                self.send_webcam_stop(device_id, message)
                return False
            tw = webcam.get_width()
            th = webcam.get_height()
            csc = csc_spec.codec_class()
            csc.init_context(w, h, rgb_pixel_format, tw, th, src_format, typedict())
            image = csc.convert_image(bgrx_image)
            webcam.push_image(image)
            # tell the client all is good:
            self.send_webcam_ack(device_id, frame_no)
            return True
        except Exception as e:
            log("error on %ix%i frame %i using encoding %s", w, h, frame_no, encoding, exc_info=True)
            log.error("Error processing webcam frame:")
            msg = str(e)
            if not msg:
                msg = "unknown error"
                log.error(f" {webcam} error", exc_info=True)
            log.error(" %s", msg)
            self.send_webcam_stop(device_id, msg)
            self.stop_virtual_webcam(device_id)
            return False
        finally:
            free()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.net.file_transfer import FileTransferHandler
from xpra.log import Logger

log = Logger()


class FakeClient(FileTransferHandler):
    def __init__(self):
        FileTransferHandler.__init__(self)
        self.mmap = None
        self.session_name = ""
        self._focused = None
        self.readonly = False
        self.windows_enabled = True
        self.can_scale = True
        self.start_new_commands = True
        self.server_bell = True
        self.client_supports_bell = True
        self.notifications_enabled = True
        self.client_supports_notifications = True
        self.cursors_enabled = True
        self.server_cursors = True
        self.client_supports_cursors = True
        self.server_clipboard = True
        self.client_supports_clipboard = True
        self.client_clipboard_direction = "both"
        self.quality = 80
        self.speed = 80
        self.encoding = "png"
        self.server_encodings_with_quality = []
        self.server_encodings_with_speed = []
        self.speaker_allowed = True
        self.speaker_enabled = True
        self.microphone_allowed = True
        self.microphone_enabled = True
        self.server_audio_send = True
        self.server_audio_receive = True
        self.server_readonly = False
        self.bell_enabled = True
        self.webcam_forwarding = True
        self.webcam_device = None
        self.server_virtual_video_devices = 0
        self.client_supports_opengl = False
        self.title = "test"
        self.keyboard_helper = None
        self.clipboard_helper = None
        self._id_to_window = {}
        self._window_to_id = {}
        self.server_window_decorations = False
        self.server_window_frame_extents = False
        self.encoding_defaults = {}
        self.xscale = 1
        self.yscale = 1
        self.log_events = True
        self.handshake_callbacks = []

    def server_ok(self):
        return True

    def log(self, *args):
        if self.log_events:
            log.info(*args)

    def connect(self, *args):
        pass

    def get_image(self, *_args):
        return None

    def get_encodings(self):
        return ["png"]

    def show_start_new_command(self):
        pass

    def show_file_upload(self):
        pass

    def after_handshake(self, cb):
        self.handshake_callbacks.append(cb)

    def fire_handshake_callbacks(self):
        cbs = self.handshake_callbacks
        self.handshake_callbacks = []
        for callback in cbs:
            callback()

    def control_refresh(self, *args, **_kwargs):
        self.log("control_refresh(%s)", args)

    def send_refresh(self, *args):
        self.log("send_refresh(%s)", args)

    def send_refresh_all(self, *args):
        self.log("send_refresh_all(%s)", args)

    def send(self, *args):
        self.log("send(%s)", args)

    def send_positional(self, *args):
        self.log("send_positional(%s)", args)

    def update_focus(self, *args):
        self.log("update_focus(%s)", args)

    def quit(self, *args):
        self.log("quit(%s)", args)

    def handle_key_action(self, *args):
        self.log("handle_key_action(%s)", args)
        return False

    def send_mouse_position(self, *args):
        self.log("send_mouse_position(%s)", args)

    def send_button(self, *args):
        self.log("send_button%s", args)

    def send_configure_event(self, skip_geometry):
        self.log("send_configure_event(%s)", skip_geometry)

    def window_close_event(self, *args):
        self.log("window_close_event%s", args)

    def mask_to_names(self, *_args):
        return []

    def get_current_modifiers(self, *_args):
        return []

    def get_mouse_position(self):
        return 0, 0

    def request_frame_extents(self, window):
        pass

    def get_window_frame_sizes(self) -> dict[str, Any]:
        return {}

    def fsx(self, v):
        return v

    def fsy(self, v):
        return v

    def sx(self, v):
        return v

    def sy(self, v):
        return v

    def srect(self, x, y, w, h):
        return x, y, w, h

    def sp(self, x, y):
        return x, y

    def cx(self, v):
        return v

    def cy(self, v):
        return v

    def crect(self, x, y, w, h):
        return x, y, w, h

    def cp(self, x, y):
        return x, y



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.server.proxy import queue_scheduler

QueueScheduler = queue_scheduler.QueueScheduler


class QueueSchedulerTest(unittest.TestCase):

    def test_idle_timeout(self):
        qs = QueueScheduler()
        calls = []
        def idle_add(arg):
            calls.append(arg)
        qs.idle_add(idle_add, True)
        def timeout_add(arg):
            calls.append(arg)
        qs.timeout_add(100, timeout_add, True)
        qs.timeout_add(500, qs.stop)
        qs.run()
        assert len(calls)==2

    def test_idle_repeat(self):
        qs = QueueScheduler()
        calls = []
        def idle_add(arg):
            calls.append(arg)
            return len(calls)<10
        qs.idle_add(idle_add, True)
        qs.timeout_add(500, qs.stop)
        qs.run()
        assert len(calls)==10

    def test_idle_cancel(self):
        qs = QueueScheduler()
        calls = []
        def idle_add():
            calls.append(True)
            return True
        tid = qs.idle_add(idle_add)
        copy = []
        def cancel():
            qs.source_remove(tid)
            copy[:] = list(calls)
            return False
        qs.timeout_add(500, cancel)
        qs.timeout_add(1000, qs.stop)
        qs.run()
        assert len(calls)==len(copy), "idle_add continued to run!"

    def test_timeout_cancel(self):
        qs = QueueScheduler()
        calls = []
        def timeout_add():
            calls.append(True)
            return True
        tid = qs.timeout_add(10, timeout_add)
        copy = []
        def cancel():
            qs.source_remove(tid)
            copy[:] = list(calls)
            return False
        qs.timeout_add(500, cancel)
        qs.timeout_add(1000, qs.stop)
        qs.run()
        assert len(calls)==len(copy), "idle_add continued to run!"

    def test_source_remove(self):
        qs = QueueScheduler()
        calls = []
        def timeout_add(arg):
            calls.append(arg)
            return True
        t = qs.timeout_add(100, timeout_add, True)
        qs.source_remove(t)
        qs.timeout_add(500, qs.stop)
        qs.run()
        assert not calls

    def test_invalid_remove(self):
        qs = QueueScheduler()
        qs.source_remove(-1)

    def test_idle_raises_exception(self):
        qs = QueueScheduler()
        def raise_exception():
            raise Exception("test scheduler error handling")
        from unit.test_util import silence_error
        with silence_error(queue_scheduler):
            qs.idle_add(raise_exception)
            qs.timeout_add(500, qs.stop)
            qs.run()

    def test_stop_queue(self):
        qs = QueueScheduler()
        qs.idle_add(qs.stop_main_queue)
        qs.run()

    def test_timer_repeat(self):
        times = list(range(10))
        qs = QueueScheduler()
        def timer_fn():
            times.pop()
            if not times:
                qs.stop()
                return False
            return True
        qs.timeout_add(1, timer_fn)
        qs.timeout_add(2000, qs.stop)
        qs.run()
        assert not times, "items remain in list: %s" % (times,)

def main():
    unittest.main()

if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.os_util import gi_import
from xpra.gtk.error import XError, xsync
from xpra.gtk.info import get_screen_sizes
from xpra.x11.gtk.bindings import add_event_receiver, remove_event_receiver
from xpra.x11.desktop.model_base import DesktopModelBase
from xpra.x11.bindings.randr import RandRBindings
from xpra.x11.bindings.window import X11WindowBindings
from xpra.log import Logger

X11Window = X11WindowBindings()
RandR = RandRBindings()

GObject = gi_import("GObject")

geomlog = Logger("server", "window", "geometry")
screenlog = Logger("screen")

# we should also figure out what the potential increments are,
# rather than hard-coding them here:
INC_VALUES = (16, 32, 64, 128, 256)


def get_legacy_size_hints(screen_sizes) -> dict[str, tuple[int, int]]:
    size_hints = {}
    # find the maximum size supported:
    max_size: dict[int, tuple[int, int]] = {}
    for tw, th in screen_sizes:
        max_size[tw * th] = (tw, th)
    max_pixels = sorted(max_size.keys())[-1]
    size_hints["maximum-size"] = max_size[max_pixels]
    # find the best increment we can use:
    inc_hits = {}
    for inc in INC_VALUES:
        hits = 0
        for tsize in screen_sizes:
            tw, th = tsize
            if (tw + inc, th + inc) in screen_sizes:
                hits += 1
        inc_hits[inc] = hits
    screenlog("size increment hits: %s", inc_hits)
    max_hits = max(inc_hits.values())
    if max_hits > 16:
        # find the first increment value matching the max hits
        for inc in INC_VALUES:
            if inc_hits[inc] == max_hits:
                break
        # TODO: also get these values from the screen sizes:
        size_hints |= {
            "base-size": (640, 640),
            "minimum-size": (640, 640),
            "increment": (128, 128),
            "minimum-aspect-ratio": (1, 3),
            "maximum-aspect-ratio": (3, 1),
        }
    return size_hints


class ScreenDesktopModel(DesktopModelBase):
    """
    A desktop model covering the entire screen as a single window.
    """
    __gsignals__ = dict(DesktopModelBase.__common_gsignals__)
    _property_names = DesktopModelBase._property_names + ["xid"]
    _dynamic_property_names = ["size-hints", "title", "icons"]

    def __init__(self, resize=True, resize_exact=False):
        super().__init__()
        self.can_resize = resize
        self.resize_exact = resize_exact

    def __repr__(self):
        return f"ScreenDesktopModel({self.xid:x})"

    def setup(self) -> None:
        super().setup()
        add_event_receiver(self.xid, self)
        self.update_size_hints()

    def unmanage(self, _exiting=False) -> None:
        remove_event_receiver(self.xid, self)

    def get_geometry(self) -> tuple[int, int, int, int]:
        w, h = self.get_dimensions()
        return 0, 0, w, h

    def get_dimensions(self) -> tuple[int, int]:
        with xsync:
            return RandR.get_screen_size()

    def get_property(self, prop: str) -> Any:
        if prop == "xid":
            return self.xid
        return super().get_property(prop)

    def do_resize(self) -> None:
        self.resize_timer = 0
        rw, rh = self.resize_value
        try:
            with xsync:
                ow, oh = RandR.get_screen_size()
            if ow == rw and oh == rh:
                return
            with xsync:
                if RandR.is_dummy16() and (rw, rh) not in get_screen_sizes():
                    RandR.add_screen_size(rw, rh)
            with xsync:
                if not RandR.set_screen_size(rw, rh):
                    geomlog.warn("Warning: failed to resize vfb")
                    return
            with xsync:
                w, h = RandR.get_screen_size()
                geomlog(f"wanted {rw}x{rh} - got {w}x{h}")
            self._screen_size_changed()
        except Exception as e:
            geomlog("do_resize() %ix%i", rw, rh, exc_info=True)
            geomlog.error("Error: failed to resize desktop display to %ix%i:", rw, rh)
            geomlog.error(" %s", str(e) or type(e))

    def do_x11_configure_event(self, event) -> None:
        self._screen_size_changed()

    def _screen_size_changed(self):
        w, h = self.get_dimensions()
        screenlog("screen size changed: new size %ix%i", w, h)
        self.invalidate_pixmap()
        self.update_size_hints()
        self.emit("resized")

    def update_size_hints(self) -> None:
        w, h = self.get_dimensions()
        screenlog("screen dimensions: %ix%i", w, h)
        size_hints: dict[str, tuple[int, int]] = {}

        def use_fixed_size() -> None:
            size = w, h
            size_hints.update({
                "maximum-size": size,
                "minimum-size": size,
                "base-size": size,
            })

        if self.can_resize:
            if self.resize_exact:
                # assume resize_exact is enabled
                # no size restrictions
                size_hints = {}
            else:
                try:
                    with xsync:
                        screen_sizes = RandR.get_xrr_screen_sizes()
                except XError:
                    screenlog("failed to query screen sizes", exc_info=True)
                else:
                    if not screen_sizes:
                        use_fixed_size()
                    else:
                        size_hints |= get_legacy_size_hints(screen_sizes)
        else:
            use_fixed_size()
        screenlog("size-hints=%s", size_hints)
        self._updateprop("size-hints", size_hints)


GObject.type_register(ScreenDesktopModel)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import ctypes
import struct
from weakref import WeakValueDictionary
from typing import Any
from collections.abc import Callable, Sequence

import Quartz
from Quartz import kCGDisplaySetModeFlag
import Quartz.CoreGraphics as CG
from Quartz.CoreGraphics import CGDisplayRegisterReconfigurationCallback, CGDisplayRemoveReconfigurationCallback
from AppKit import NSObject, NSAppleEventManager, NSScreen, NSBeep, NSApp

from xpra.os_util import gi_import
from xpra.common import roundup
from xpra.exit_codes import ExitValue
from xpra.util.env import envint, envbool
from xpra.util.io import CaptureStdErr
from xpra.platform.darwin import get_OSXApplication
from xpra.log import Logger

log = Logger("osx", "events")
workspacelog = Logger("osx", "events", "workspace")
mouselog = Logger("osx", "events", "mouse")
notifylog = Logger("osx", "notify")

GLib = gi_import("GLib")

OSX_FOCUS_WORKAROUND = envint("XPRA_OSX_FOCUS_WORKAROUND", 2000)
SLEEP_HANDLER = envbool("XPRA_OSX_SLEEP_HANDLER", True)
EVENT_LISTENER = envbool("XPRA_OSX_EVENT_LISTENER", True)
OSX_WHEEL_MULTIPLIER = envint("XPRA_OSX_WHEEL_MULTIPLIER", 100)
OSX_WHEEL_PRECISE_MULTIPLIER = envint("XPRA_OSX_WHEEL_PRECISE_MULTIPLIER", 1)
OSX_WHEEL_DIVISOR = envint("XPRA_OSX_WHEEL_DIVISOR", 10)
WHEEL = envbool("XPRA_WHEEL", True)
NATIVE_NOTIFIER = envbool("XPRA_OSX_NATIVE_NOTIFIER", True)
SUBPROCESS_NOTIFIER = envbool("XPRA_OSX_SUBPROCESS_NOTIFIER", False)

ALPHA = {
    CG.kCGImageAlphaNone: "AlphaNone",
    CG.kCGImageAlphaPremultipliedLast: "PremultipliedLast",
    CG.kCGImageAlphaPremultipliedFirst: "PremultipliedFirst",
    CG.kCGImageAlphaLast: "Last",
    CG.kCGImageAlphaFirst: "First",
    CG.kCGImageAlphaNoneSkipLast: "SkipLast",
    CG.kCGImageAlphaNoneSkipFirst: "SkipFirst",
}


# if there is an easier way of doing this, I couldn't find it:
def nodbltime() -> float:
    return -1


GetDblTime: Callable = nodbltime
try:
    Carbon_ctypes = ctypes.CDLL("/System/Library/Frameworks/Carbon.framework/Carbon")
    GetDblTime = Carbon_ctypes.GetDblTime
except Exception:
    log("GetDblTime not found", exc_info=True)


def do_init() -> None:
    osxapp = get_OSXApplication()
    log("do_init() osxapp=%s", osxapp)
    if not osxapp:
        return  # not much else we can do here
    from xpra.platform.paths import get_icon
    from xpra.platform.gui import get_default_icon
    filename = get_default_icon()
    icon = get_icon(filename)
    log("do_init() icon=%s", icon)
    if icon:
        osxapp.set_dock_icon_pixbuf(icon)
    from xpra.platform.darwin.menu import getOSXMenuHelper
    mh = getOSXMenuHelper(None)
    log("do_init() menu helper=%s", mh)
    osxapp.set_dock_menu(mh.build_dock_menu())
    import warnings
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=".*invalid cast from 'GtkMenuBar'")
        with CaptureStdErr():
            osxapp.set_menu_bar(mh.rebuild())


def do_ready() -> None:
    osxapp = get_OSXApplication()
    if osxapp:
        log("%s()", osxapp.ready)
        osxapp.ready()


def get_clipboard_native_class() -> str:
    return "xpra.platform.darwin.clipboard.OSXClipboardProtocolHelper"


def get_native_notifier_classes() -> list[type]:
    v: list[type] = []
    if NATIVE_NOTIFIER:
        from Foundation import NSUserNotificationCenter
        if NSUserNotificationCenter.defaultUserNotificationCenter():
            from xpra.platform.darwin.notifier import OSX_Notifier
            v.append(OSX_Notifier)
    notifylog("get_native_notifier_classes()=%s", v)
    return v


def get_native_tray_menu_helper_class() -> Callable | None:
    if get_OSXApplication():
        from xpra.platform.darwin.menu import getOSXMenuHelper
        return getOSXMenuHelper
    return None


def get_native_tray_classes() -> list[type]:
    if get_OSXApplication():
        from xpra.platform.darwin.tray import OSXTray
        return [OSXTray]
    return []


def system_bell(*_args) -> bool:
    NSBeep()
    return True


def _sizetotuple(s) -> tuple[int, int]:
    return int(s.width), int(s.height)


def _recttotuple(r) -> tuple[int, int, int, int]:
    return int(r.origin.x), int(r.origin.y), int(r.size.width), int(r.size.height)


def get_double_click_time() -> int:
    try:
        # what are ticks? just an Apple retarded way of measuring elapsed time.
        # They must have considered gigaparsecs divided by teapot too, which is just as useful.
        # (but still call it "Time" you see)
        MS_PER_TICK = 1000.0 / 60
        v = GetDblTime()
        if v > 0:
            return int(v * MS_PER_TICK)
    except Exception:
        pass
    return -1


def get_window_min_size() -> tuple[int, int]:
    # roughly enough to see the window buttons:
    return 120, 1


# def get_window_max_size():
#    return 2**15-1, 2**15-1


def get_window_frame_sizes() -> dict[str, Any]:
    # use a hard-coded window position and size:
    return get_window_frame_size(20, 20, 100, 100)


def get_window_frame_size(x: int, y: int, w: int, h: int) -> dict[str, Any]:
    try:
        cr = Quartz.NSMakeRect(x, y, w, h)
        mask = Quartz.NSTitledWindowMask | Quartz.NSClosableWindowMask | Quartz.NSMiniaturizableWindowMask | Quartz.NSResizableWindowMask
        wr = Quartz.NSWindow.pyobjc_classMethods.frameRectForContentRect_styleMask_(cr, mask)
        dx = int(wr[0][0] - cr[0][0])
        dy = int(wr[0][1] - cr[0][1])
        dw = int(wr[1][0] - cr[1][0])
        dh = int(wr[1][1] - cr[1][1])
        # note: we assume that the title bar is at the top
        # dx, dy and dw are usually 0
        # dh is usually 22 on my 10.5.x system
        return {
            "offset": (dx + dw // 2, dy + dh),
            "frame": (dx + dw // 2, dx + dw // 2, dy + dh, dy),
        }
    except Exception:
        log("failed to query frame size using Quartz, using hardcoded value", exc_info=True)
        return {  # left, right, top, bottom:
            "offset": (0, 22),
            "frame": (0, 0, 22, 0),
        }


def get_workarea() -> tuple[int, int, int, int] | None:
    w = get_workareas()
    if w and len(w) == 1:
        return w[0]
    return None


# per monitor workareas (assuming a single screen)
def get_workareas() -> Sequence[tuple[int, int, int, int]]:
    workareas = []
    screens = NSScreen.screens()
    for screen in screens:
        log("get_workareas() testing screen %s", screen)
        frame = screen.frame()
        visibleFrame = screen.visibleFrame()
        log(" frame=%s, visibleFrame=%s", frame, visibleFrame)
        log(" backingScaleFactor=%s", screen.backingScaleFactor())
        x = int(visibleFrame.origin.x - frame.origin.x)
        y = int((frame.size.height - visibleFrame.size.height) - (visibleFrame.origin.y - frame.origin.y))
        w = int(visibleFrame.size.width)
        h = int(visibleFrame.size.height)
        workareas.append((x, y, w, h))
    log("get_workareas()=%s", workareas)
    return workareas


def get_vrefresh() -> int:
    vrefresh = []
    try:
        err, active_displays, no = CG.CGGetActiveDisplayList(99, None, None)
        log("get_vrefresh() %i active displays: %s (err=%i)", no, active_displays, err)
        if err == 0 and no > 0:
            for adid in active_displays:
                mode = CG.CGDisplayCopyDisplayMode(adid)
                v = int(CG.CGDisplayModeGetRefreshRate(mode))
                log("get_vrefresh() refresh-rate(%#x)=%i", adid, v)
                if v > 0:
                    vrefresh.append(v)
    except Exception:
        log("failed to query vrefresh for active displays: %s", exc_info=True)
    log("get_vrefresh() found %s", vrefresh)
    if len(vrefresh) > 0:
        return min(vrefresh)
    return -1


def get_display_icc_info() -> dict[Any, dict]:
    info = {}
    try:
        err, active_displays, no = CG.CGGetActiveDisplayList(99, None, None)
        if err == 0 and no > 0:
            for i, adid in enumerate(active_displays):
                info[i] = get_colorspace_info(CG.CGDisplayCopyColorSpace(adid))
    except Exception as e:
        log("failed to query colorspace for active displays: %s", e)
    return info


def get_icc_info() -> dict:
    # maybe we shouldn't return anything if there's more than one display?
    info = {}
    try:
        did = CG.CGMainDisplayID()
        info = get_colorspace_info(CG.CGDisplayCopyColorSpace(did))
    except Exception as e:
        log("failed to query colorspace for main display: %s", e)
    return info


def get_colorspace_info(cs) -> dict[str, Any]:
    MODELS: dict[Any, str] = {
        CG.kCGColorSpaceModelUnknown: "unknown",
        CG.kCGColorSpaceModelMonochrome: "monochrome",
        CG.kCGColorSpaceModelRGB: "RGB",
        CG.kCGColorSpaceModelCMYK: "CMYK",
        CG.kCGColorSpaceModelLab: "lab",
        CG.kCGColorSpaceModelDeviceN: "DeviceN",
        CG.kCGColorSpaceModelIndexed: "indexed",
        CG.kCGColorSpaceModelPattern: "pattern",
    }

    # base = CGColorSpaceGetBaseColorSpace(cs)
    # color_table = CGColorSpaceGetColorTable(cs)

    def tomodelstr(v) -> str:
        return MODELS.get(v, "unknown")

    defs = (
        ("name", "CGColorSpaceCopyName", str),
        ("icc-profile", "CGColorSpaceCopyICCProfile", str),
        ("icc-data", "CGColorSpaceCopyICCData", str),
        ("components", "CGColorSpaceGetNumberOfComponents", int),
        ("supports-output", "CGColorSpaceSupportsOutput", bool),
        ("model", "CGColorSpaceGetModel", tomodelstr),
        ("wide-gamut", "CGColorSpaceIsWideGamutRGB", bool),
        ("color-table-count", "CGColorSpaceGetColorTableCount", int),
    )
    return _call_CG_conv(defs, cs)


def get_display_mode_info(mode) -> dict[str, Any]:
    defs = (
        ("width", "CGDisplayModeGetWidth", int),
        ("height", "CGDisplayModeGetHeight", int),
        ("pixel-encoding", "CGDisplayModeCopyPixelEncoding", str),
        ("vrefresh", "CGDisplayModeGetRefreshRate", int),
        ("io-flags", "CGDisplayModeGetIOFlags", int),
        ("id", "CGDisplayModeGetIODisplayModeID", int),
        ("usable-for-desktop", "CGDisplayModeIsUsableForDesktopGUI", bool),
    )
    return _call_CG_conv(defs, mode)


def get_display_modes_info(modes) -> dict[int, Any]:
    return {i: get_display_mode_info(mode) for i, mode in enumerate(modes)}


def _call_CG_conv(defs: Sequence[tuple[str, str, Callable]], argument) -> dict[str, Any]:
    # utility for calling functions on CG with an argument,
    # then convert the return value using another function
    # missing functions are ignored, and None values are skipped
    info = {}
    for prop_name, fn_name, conv in defs:
        fn = getattr(CG, fn_name, None)
        if fn:
            try:
                v = fn(argument)
            except Exception as e:
                log("function %s failed: %s", fn_name, e)
                del e
                continue
            if v is not None:
                info[prop_name] = conv(v)
            else:
                log("%s is not set", prop_name)
        else:
            log("function %s does not exist", fn_name)
    return info


def get_display_info(did) -> dict[str, Any]:
    defs = (
        ("height", "CGDisplayPixelsHigh", int),
        ("width", "CGDisplayPixelsWide", int),
        ("bounds", "CGDisplayBounds", _recttotuple),
        ("active", "CGDisplayIsActive", bool),
        ("asleep", "CGDisplayIsAsleep", bool),
        ("online", "CGDisplayIsOnline", bool),
        ("main", "CGDisplayIsMain", bool),
        ("builtin", "CGDisplayIsBuiltin", bool),
        ("in-mirror-set", "CGDisplayIsInMirrorSet", bool),
        ("always-in-mirror-set", "CGDisplayIsAlwaysInMirrorSet", bool),
        ("in-hw-mirror-set", "CGDisplayIsInHWMirrorSet", bool),
        ("mirrors-display", "CGDisplayMirrorsDisplay", int),
        ("stereo", "CGDisplayIsStereo", bool),
        ("primary", "CGDisplayPrimaryDisplay", bool),
        ("unit-number", "CGDisplayUnitNumber", int),
        ("vendor", "CGDisplayVendorNumber", int),
        ("model", "CGDisplayModelNumber", int),
        ("serial", "CGDisplaySerialNumber", int),
        ("service-port", "CGDisplayIOServicePort", int),
        ("size", "CGDisplayScreenSize", _sizetotuple),
        ("rotation", "CGDisplayRotation", int),
        ("colorspace", "CGDisplayCopyColorSpace", get_colorspace_info),
        ("opengl-acceleration", "CGDisplayUsesOpenGLAcceleration", bool),
        ("mode", "CGDisplayCopyDisplayMode", get_display_mode_info),
    )
    info = _call_CG_conv(defs, did)
    try:
        modes = CG.CGDisplayCopyAllDisplayModes(did, None)
        info["modes"] = get_display_modes_info(modes)
    except Exception as e:
        log("failed to query display modes: %s", e)
    return info


def get_displays_info() -> dict[str, Any]:
    did = CG.CGMainDisplayID()
    info: dict[str, Any] = {
        "main": get_display_info(did),
    }
    err, active_displays, no = CG.CGGetActiveDisplayList(99, None, None)
    if err == 0 and no > 0:
        for i, adid in enumerate(active_displays):
            info.setdefault("active", {})[i] = get_display_info(adid)
    err, online_displays, no = CG.CGGetOnlineDisplayList(99, None, None)
    if err == 0 and no > 0:
        for i, odid in enumerate(online_displays):
            info.setdefault("online", {})[i] = get_display_info(odid)
    return info


def get_info() -> dict[str, Any]:
    from xpra.platform.gui import get_info_base
    i = get_info_base()
    with log.trap_error("Error: OSX get_display_info failed"):
        i["displays"] = get_displays_info()
    return i


# keep track of the window object for each view
VIEW_TO_WINDOW: WeakValueDictionary[int, Any] = WeakValueDictionary()


def get_nsview(window) -> int:
    try:
        from xpra.platform.darwin.gdk3_bindings import get_nsview_ptr
    except ImportError:
        return 0
    try:
        return get_nsview_ptr(window.get_window())
    except Exception:
        return 0


def add_window_hooks(window) -> None:
    if WHEEL:
        nsview = get_nsview(window)
        if nsview:
            VIEW_TO_WINDOW[nsview] = window


def remove_window_hooks(window) -> None:
    if WHEEL:
        nsview = get_nsview(window)
        if nsview:
            VIEW_TO_WINDOW.pop(nsview, None)


def get_CG_imagewrapper(rect=None):
    from xpra.codecs.image import ImageWrapper
    assert CG, "cannot capture without Quartz.CoreGraphics"
    if rect is None:
        x = 0
        y = 0
        region = CG.CGRectInfinite
    else:
        x, y, w, h = rect
        region = CG.CGRectMake(x, y, roundup(w, 2), roundup(h, 2))
    image = CG.CGWindowListCreateImage(region,
                                       CG.kCGWindowListOptionOnScreenOnly,
                                       CG.kCGNullWindowID,
                                       CG.kCGWindowImageNominalResolution)  # CG.kCGWindowImageDefault)
    width = CG.CGImageGetWidth(image)
    height = CG.CGImageGetHeight(image)
    bpc = CG.CGImageGetBitsPerComponent(image)
    bpp = CG.CGImageGetBitsPerPixel(image)
    rowstride = CG.CGImageGetBytesPerRow(image)
    alpha = CG.CGImageGetAlphaInfo(image)
    alpha_str = ALPHA.get(alpha, alpha)
    log("get_CG_imagewrapper(..) image size: %sx%s, bpc=%s, bpp=%s, rowstride=%s, alpha=%s", width, height, bpc, bpp,
        rowstride, alpha_str)
    prov = CG.CGImageGetDataProvider(image)
    argb = CG.CGDataProviderCopyData(prov)
    return ImageWrapper(x, y, width, height, argb, "BGRX", 24, rowstride)


def take_screenshot() -> tuple[int, int, str, int, bytes]:
    log("grabbing screenshot")
    from PIL import Image
    from io import BytesIO
    image = get_CG_imagewrapper()
    w = image.get_width()
    h = image.get_height()
    img = Image.frombuffer("RGB", (w, h), image.get_pixels(), "raw", image.get_pixel_format(), image.get_rowstride())
    buf = BytesIO()
    img.save(buf, "PNG")
    data = buf.getvalue()
    buf.close()
    return w, h, "png", image.get_rowstride(), data


def force_focus(duration=2000) -> None:
    enable_focus_workaround()
    GLib.timeout_add(duration, disable_focus_workaround)


__osx_open_signal = False
# We have to wait for the main loop to be running
# to get the NSApplicationOpenFile signal,
# or the GURLHandler.
OPEN_SIGNAL_WAIT = envint("XPRA_OSX_OPEN_SIGNAL_WAIT", 500)


def add_open_handlers(open_file_cb: Callable, open_url_cb: Callable) -> None:

    def open_URL(url) -> bool:
        global __osx_open_signal
        __osx_open_signal = True
        log("open_URL(%s)", url)
        GLib.idle_add(open_url_cb, url)
        return True

    def open_file(_, filename: str, *args) -> bool:
        global __osx_open_signal
        __osx_open_signal = True
        log("open_file(%s, %s)", filename, args)
        GLib.idle_add(open_file_cb, filename)
        return True

    register_file_handler(open_file)
    register_URL_handler(open_URL)


def wait_for_open_handlers(show_cb: Callable,
                           open_file_cb: Callable,
                           open_url_cb: Callable,
                           delay: int = OPEN_SIGNAL_WAIT) -> None:
    add_open_handlers(open_file_cb, open_url_cb)

    def may_show():
        global __osx_open_signal
        log("may_show() osx open signal=%s", __osx_open_signal)
        if not __osx_open_signal:
            force_focus()
            show_cb()

    GLib.timeout_add(delay, may_show)


def register_file_handler(handler: Callable) -> None:
    log("register_file_handler(%s)", handler)
    try:
        get_OSXApplication().connect("NSApplicationOpenFile", handler)
    except Exception as e:
        log.error("Error: cannot handle file associations:")
        log.estr(e)


def register_URL_handler(handler: Callable) -> None:
    log("register_URL_handler(%s)", handler)

    class GURLHandler(NSObject):
        # noinspection PyMethodMayBeStatic
        def handleEvent_withReplyEvent_(self, event, reply_event):
            log("GURLHandler.handleEvent(%s, %s)", event, reply_event)
            url = event.descriptorForKeyword_(four_char_to_int(b'----')).stringValue()
            log("URL=%s", url)
            handler(url.encode())

    # A helper to make struct since cocoa headers seem to make
    # it impossible to use kAE*
    def four_char_to_int(code: bytes) -> int:
        return struct.unpack(b'>l', code)[0]

    urlh = GURLHandler.alloc()
    urlh.init()
    urlh.retain()
    manager = NSAppleEventManager.sharedAppleEventManager()
    manager.setEventHandler_andSelector_forEventClass_andEventID_(
        urlh, 'handleEvent:withReplyEvent:',
        four_char_to_int(b'GURL'), four_char_to_int(b'GURL')
    )


def disable_focus_workaround() -> None:
    NSApp.activateIgnoringOtherApps_(False)


def enable_focus_workaround() -> None:
    NSApp.activateIgnoringOtherApps_(True)


def can_access_display() -> bool:
    # see: https://stackoverflow.com/a/11511419/428751
    d = Quartz.CGSessionCopyCurrentDictionary()
    if not d:
        return False
    kCGSSessionOnConsoleKey = d.get("kCGSSessionOnConsoleKey", 0)
    log("kCGSSessionOnConsoleKey=%s", kCGSSessionOnConsoleKey)
    if kCGSSessionOnConsoleKey == 0:
        # GUI session doesn't own the console, or the console's screens are asleep
        return False
    CGSSessionScreenIsLocked = d.get("CGSSessionScreenIsLocked", 0)
    log("CGSSessionScreenIsLocked=%s", CGSSessionScreenIsLocked)
    if CGSSessionScreenIsLocked:
        # screen is locked
        return False
    return True


class ClientExtras:
    def __init__(self, client=None, opts=None):
        if OSX_FOCUS_WORKAROUND and client:
            def first_ui_received(*_args):
                enable_focus_workaround()
                GLib.timeout_add(OSX_FOCUS_WORKAROUND, disable_focus_workaround)

            client.connect("first-ui-received", first_ui_received)
        log("ClientExtras.__init__(%s, %s)", client, opts)
        self.client = client
        self.event_loop_started = False
        self.check_display_timer = 0
        self.display_is_asleep = False
        if opts and client:
            swap_keys = opts.swap_keys
            log("setting swap_keys=%s using %s", swap_keys, client.keyboard_helper)
            if client.keyboard_helper and client.keyboard_helper.keyboard:
                log("%s.swap_keys=%s", client.keyboard_helper.keyboard, swap_keys)
                client.keyboard_helper.keyboard.swap_keys = swap_keys
        if client:
            self.check_display_timer = GLib.timeout_add(60 * 1000, self.check_display)

    def cleanup(self) -> None:
        cdt = self.check_display_timer
        if cdt:
            GLib.source_remove(cdt)
            self.check_display_timer = 0
        try:
            r = CGDisplayRemoveReconfigurationCallback(self.display_change, self)
        except ValueError as e:
            log("CGDisplayRemoveReconfigurationCallback: %s", e)
            # if we exit from a signal, this may fail
            r = 1
        if r != 0:
            # don't bother logging this as a warning since we are terminating anyway:
            log("failed to unregister display reconfiguration callback")
        self.client = None

    def ready(self) -> None:
        if EVENT_LISTENER:
            with log.trap_error("Error setting up OSX event listener"):
                self.setup_event_listener()

    def setup_event_listener(self) -> None:
        log(f"setup_event_listener() client={self.client}")
        if self.client:
            from xpra.platform.darwin.events import get_app_delegate
            delegate = get_app_delegate()
            log(f"setup_event_listener() delegate={delegate}")
            delegate.add_handler("suspend", self.client.suspend)
            delegate.add_handler("resume", self.client.resume)
            delegate.add_handler("deiconify", self.client.deiconify_windows)
        r = CGDisplayRegisterReconfigurationCallback(self.display_change, self)
        if r != 0:
            log.warn("Warning: failed to register display reconfiguration callback")

    def display_change(self, display, flags, userinfo) -> None:
        log("display_change%s", (display, flags, userinfo))
        c = self.client
        # The display mode has changed
        # opengl windows may need to be re-created since the GPU may have changed:
        if (flags & kCGDisplaySetModeFlag) and c and c.opengl_enabled:
            c.reinit_windows()

    def check_display(self) -> bool:
        log("check_display()")
        try:
            c = self.client
            asleep = None
            if not can_access_display():
                asleep = True
            else:
                did = CG.CGMainDisplayID()
                log("check_display() CGMainDisplayID()=%#x", did)
                if did and c:
                    asleep = bool(CG.CGDisplayIsAsleep(did))
                    log("check_display() CGDisplayIsAsleep(%#x)=%s", did, asleep)
            if c and asleep is not None and self.display_is_asleep != asleep:
                self.display_is_asleep = asleep
                if asleep:
                    c.suspend()
                else:
                    c.resume()
            return True
        except Exception:
            log.error("Error checking display sleep status", exc_info=True)
            self.check_display_timer = 0
            return False

    def run(self) -> ExitValue:
        # this is for running standalone
        log("starting console event loop")
        self.event_loop_started = True
        import PyObjCTools.AppHelper as AppHelper
        AppHelper.runConsoleEventLoop(installInterrupt=True)
        # when running from the GTK main loop, we rely on another part of the code
        # to run the event loop for us
        return 0

    def stop(self) -> None:
        if self.event_loop_started:
            self.event_loop_started = False
            import PyObjCTools.AppHelper as AppHelper
            AppHelper.stopEventLoop()


def main() -> ExitValue:
    from xpra.platform import program_context
    with program_context("OSX Extras"):
        log.enable_debug()
        ce = ClientExtras(None, None)
        ce.check_display()
        ce.ready()
        return ce.run()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from libc.stdint cimport uint8_t, uint32_t, uint64_t, uintptr_t   #pylint: disable=syntax-error

DEF AVIF_PLANE_COUNT_YUV = 3

cdef extern from "avif/avif.h":
    int AVIF_VERSION_MAJOR
    int AVIF_VERSION_MINOR
    int AVIF_VERSION_PATCH

    int AVIF_QUANTIZER_LOSSLESS
    int AVIF_QUANTIZER_BEST_QUALITY
    int AVIF_QUANTIZER_WORST_QUALITY

    ctypedef int avifBool

    ctypedef enum avifPlanesFlags:
        AVIF_PLANES_YUV
        AVIF_PLANES_A
        AVIF_PLANES_ALL

    ctypedef enum avifChannelIndex:
        AVIF_CHAN_R
        AVIF_CHAN_G
        AVIF_CHAN_B

        AVIF_CHAN_Y
        AVIF_CHAN_U
        AVIF_CHAN_V

    ctypedef struct avifROData:
        const uint8_t * data
        size_t size

    ctypedef struct avifRWData:
        uint8_t * data
        size_t size

    ctypedef enum avifPixelFormat:
        AVIF_PIXEL_FORMAT_NONE
        AVIF_PIXEL_FORMAT_YUV444
        AVIF_PIXEL_FORMAT_YUV422
        AVIF_PIXEL_FORMAT_YUV420
        AVIF_PIXEL_FORMAT_YUV400

    ctypedef struct avifPixelFormatInfo:
        avifBool monochrome
        int chromaShiftX
        int chromaShiftY

    ctypedef enum avifChromaSamplePosition:
        AVIF_CHROMA_SAMPLE_POSITION_UNKNOWN
        AVIF_CHROMA_SAMPLE_POSITION_VERTICAL
        AVIF_CHROMA_SAMPLE_POSITION_COLOCATED

    ctypedef enum avifRange:
        AVIF_RANGE_LIMITED = 0,
        AVIF_RANGE_FULL = 1

    ctypedef enum avifColorPrimaries:
        AVIF_COLOR_PRIMARIES_UNKNOWN
        AVIF_COLOR_PRIMARIES_BT709
        AVIF_COLOR_PRIMARIES_IEC61966_2_4
        AVIF_COLOR_PRIMARIES_UNSPECIFIED
        AVIF_COLOR_PRIMARIES_BT470M
        AVIF_COLOR_PRIMARIES_BT470BG
        AVIF_COLOR_PRIMARIES_BT601
        AVIF_COLOR_PRIMARIES_SMPTE240
        AVIF_COLOR_PRIMARIES_GENERIC_FILM
        AVIF_COLOR_PRIMARIES_BT2020
        AVIF_COLOR_PRIMARIES_XYZ
        AVIF_COLOR_PRIMARIES_SMPTE431
        AVIF_COLOR_PRIMARIES_SMPTE432
        AVIF_COLOR_PRIMARIES_EBU3213

    ctypedef enum avifTransferCharacteristics:
        AVIF_TRANSFER_CHARACTERISTICS_UNKNOWN
        AVIF_TRANSFER_CHARACTERISTICS_BT709
        AVIF_TRANSFER_CHARACTERISTICS_UNSPECIFIED
        AVIF_TRANSFER_CHARACTERISTICS_BT470M
        AVIF_TRANSFER_CHARACTERISTICS_BT470BG
        AVIF_TRANSFER_CHARACTERISTICS_BT601
        AVIF_TRANSFER_CHARACTERISTICS_SMPTE240
        AVIF_TRANSFER_CHARACTERISTICS_LINEAR
        AVIF_TRANSFER_CHARACTERISTICS_LOG100
        AVIF_TRANSFER_CHARACTERISTICS_LOG100_SQRT10
        AVIF_TRANSFER_CHARACTERISTICS_IEC61966
        AVIF_TRANSFER_CHARACTERISTICS_BT1361
        AVIF_TRANSFER_CHARACTERISTICS_SRGB
        AVIF_TRANSFER_CHARACTERISTICS_BT2020_10BIT
        AVIF_TRANSFER_CHARACTERISTICS_BT2020_12BIT
        AVIF_TRANSFER_CHARACTERISTICS_SMPTE2084
        AVIF_TRANSFER_CHARACTERISTICS_SMPTE428
        AVIF_TRANSFER_CHARACTERISTICS_HLG

    ctypedef enum avifMatrixCoefficients:
        AVIF_MATRIX_COEFFICIENTS_IDENTITY
        AVIF_MATRIX_COEFFICIENTS_BT709
        AVIF_MATRIX_COEFFICIENTS_UNSPECIFIED
        AVIF_MATRIX_COEFFICIENTS_FCC
        AVIF_MATRIX_COEFFICIENTS_BT470BG
        AVIF_MATRIX_COEFFICIENTS_BT601
        AVIF_MATRIX_COEFFICIENTS_SMPTE240
        AVIF_MATRIX_COEFFICIENTS_YCGCO
        AVIF_MATRIX_COEFFICIENTS_BT2020_NCL
        AVIF_MATRIX_COEFFICIENTS_BT2020_CL
        AVIF_MATRIX_COEFFICIENTS_SMPTE2085
        AVIF_MATRIX_COEFFICIENTS_CHROMA_DERIVED_NCL
        AVIF_MATRIX_COEFFICIENTS_CHROMA_DERIVED_CL
        AVIF_MATRIX_COEFFICIENTS_ICTCP

    ctypedef enum avifTransformationFlags:
        AVIF_TRANSFORM_NONE
        AVIF_TRANSFORM_PASP
        AVIF_TRANSFORM_CLAP
        AVIF_TRANSFORM_IROT
        AVIF_TRANSFORM_IMIR

    ctypedef struct avifPixelAspectRatioBox:
        uint32_t hSpacing
        uint32_t vSpacing

    ctypedef struct avifCleanApertureBox:
        uint32_t widthN
        uint32_t widthD
        #a fractional number which defines the exact clean aperture height, in counted pixels, of the video image
        uint32_t heightN
        uint32_t heightD
        #a fractional number which defines the horizontal offset of clean aperture centre minus (width-1)/2. Typically 0.
        uint32_t horizOffN
        uint32_t horizOffD
        #a fractional number which defines the vertical offset of clean aperture centre minus (height-1)/2. Typically 0.
        uint32_t vertOffN
        uint32_t vertOffD

    ctypedef struct avifImageRotation:
        uint8_t angle

    ctypedef struct avifImageMirror:
        # 0: Mirror about a vertical axis ("left-to-right")
        # 1: Mirror about a horizontal axis ("top-to-bottom")
        uint8_t axis

    ctypedef struct avifImage:
        uint32_t width
        uint32_t height
        uint32_t depth      #all planes must share this depth; if depth>8, all planes are uint16_t internally

        avifPixelFormat yuvFormat
        avifRange yuvRange
        avifChromaSamplePosition yuvChromaSamplePosition
        uint8_t * yuvPlanes[AVIF_PLANE_COUNT_YUV]
        uint32_t yuvRowBytes[AVIF_PLANE_COUNT_YUV]
        avifBool imageOwnsYUVPlanes

        avifRange alphaRange
        uint8_t * alphaPlane
        uint32_t alphaRowBytes
        avifBool imageOwnsAlphaPlane
        avifBool alphaPremultiplied

        #ICC Profile
        avifRWData icc

        #CICP information:
        #These are stored in the AV1 payload and used to signal YUV conversion. Additionally, if an
        #ICC profile is not specified, these will be stored in the AVIF container's `colr` box with
        #a type of `nclx`. If your system supports ICC profiles, be sure to check for the existence
        #of one (avifImage.icc) before relying on the values listed here!
        avifColorPrimaries colorPrimaries
        avifTransferCharacteristics transferCharacteristics
        avifMatrixCoefficients matrixCoefficients

        #Transformations - These metadata values are encoded/decoded when transformFlags are set
        #appropriately, but do not impact/adjust the actual pixel buffers used (images won't be
        #pre-cropped or mirrored upon decode). Basic explanations from the standards are offered in
        #comments above, but for detailed explanations, please refer to the HEIF standard (ISO/IEC
        #23008-12:2017) and the BMFF standard (ISO/IEC 14496-12:2015).
        #To encode any of these boxes, set the values in the associated box, then enable the flag in
        #transformFlags. On decode, only honor the values in boxes with the associated transform flag set.
        uint32_t transformFlags
        avifPixelAspectRatioBox pasp
        avifCleanApertureBox clap
        avifImageRotation irot
        avifImageMirror imir
        #Metadata - set with avifImageSetMetadata*() before write, check .size>0 for existence after read
        avifRWData exif
        avifRWData xmp

    ctypedef enum avifRGBFormat:
        AVIF_RGB_FORMAT_RGB
        AVIF_RGB_FORMAT_RGBA    #This is the default format set in avifRGBImageSetDefaults().
        AVIF_RGB_FORMAT_ARGB
        AVIF_RGB_FORMAT_BGR
        AVIF_RGB_FORMAT_BGRA
        AVIF_RGB_FORMAT_ABGR

    void avifGetPixelFormatInfo(avifPixelFormat format, avifPixelFormatInfo * info)
    const char * avifPixelFormatToString(avifPixelFormat format)

    void * avifAlloc(size_t size)
    void avifFree(void * p)

    void avifRWDataRealloc(avifRWData * raw, size_t newSize)
    void avifRWDataSet(avifRWData * raw, const uint8_t * data, size_t len)
    void avifRWDataFree(avifRWData * raw)

    const char * avifResultToString(avifResult result)


    avifImage * avifImageCreate(int width, int height, int depth, avifPixelFormat yuvFormat)
    avifImage * avifImageCreateEmpty()  # helper for making an image to decode into
    void avifImageCopy(avifImage * dstImage, const avifImage * srcImage, uint32_t planes)   #deep copy
    void avifImageDestroy(avifImage * image)

    void avifImageSetProfileICC(avifImage * image, const uint8_t * icc, size_t iccSize)

    #Warning: If the Exif payload is set and invalid, avifEncoderWrite() may return AVIF_RESULT_INVALID_EXIF_PAYLOAD
    void avifImageSetMetadataExif(avifImage * image, const uint8_t * exif, size_t exifSize)
    void avifImageSetMetadataXMP(avifImage * image, const uint8_t * xmp, size_t xmpSize)

    void avifImageAllocatePlanes(avifImage * image, uint32_t planes)    #Ignores any pre-existing planes
    void avifImageFreePlanes(avifImage * image, uint32_t planes)        #Ignores already-freed planes
    void avifImageStealPlanes(avifImage * dstImage, avifImage * srcImage, uint32_t planes)

    uint32_t avifRGBFormatChannelCount(avifRGBFormat format)
    avifBool avifRGBFormatHasAlpha(avifRGBFormat format)

    ctypedef enum avifChromaUpsampling:
        AVIF_CHROMA_UPSAMPLING_AUTOMATIC
        AVIF_CHROMA_UPSAMPLING_FASTEST
        AVIF_CHROMA_UPSAMPLING_BEST_QUALITY
        AVIF_CHROMA_UPSAMPLING_NEAREST
        AVIF_CHROMA_UPSAMPLING_BILINEAR

    ctypedef struct avifRGBImage:
        uint32_t width          #must match associated avifImage
        uint32_t height         #must match associated avifImage
        uint32_t depth          #legal depths [8, 10, 12, 16]. if depth>8, pixels must be uint16_t internally
        avifRGBFormat format    #all channels are always full range
        avifChromaUpsampling chromaUpsampling   # Defaults to AVIF_CHROMA_UPSAMPLING_AUTOMATIC: How to upsample non-4:4:4 UV (ignored for 444) when converting to RGB.
                                                # Unused when converting to YUV. avifRGBImageSetDefaults() prefers quality over speed.
        avifBool ignoreAlpha    #Used for XRGB formats, treats formats containing alpha (such as ARGB) as if they were
                                #RGB, treating the alpha bits as if they were all 1.
        avifBool alphaPremultiplied #indicates if RGB value is pre-multiplied by alpha. Default: false
        uint8_t * pixels
        uint32_t rowBytes

    void avifRGBImageSetDefaults(avifRGBImage * rgb, const avifImage * image)
    uint32_t avifRGBImagePixelSize(const avifRGBImage * rgb)

    # Convenience functions. If you supply your own pixels/rowBytes, you do not need to use these.
    void avifRGBImageAllocatePixels(avifRGBImage * rgb)
    void avifRGBImageFreePixels(avifRGBImage * rgb)

    #The main conversion functions
    avifResult avifImageRGBToYUV(avifImage * image, const avifRGBImage * rgb)
    avifResult avifImageYUVToRGB(const avifImage * image, avifRGBImage * rgb)

    #Premultiply handling functions.
    #(Un)premultiply is automatically done by the main conversion functions above,
    #so usually you don't need to call these. They are there for convenience.
    avifResult avifRGBImagePremultiplyAlpha(avifRGBImage * rgb)
    avifResult avifRGBImageUnpremultiplyAlpha(avifRGBImage * rgb)

    int avifFullToLimitedY(int depth, int v)
    int avifFullToLimitedUV(int depth, int v)
    int avifLimitedToFullY(int depth, int v)
    int avifLimitedToFullUV(int depth, int v)

    ctypedef enum avifCodecChoice:
        AVIF_CODEC_CHOICE_AUTO
        AVIF_CODEC_CHOICE_AOM
        AVIF_CODEC_CHOICE_DAV1D
        AVIF_CODEC_CHOICE_LIBGAV1
        AVIF_CODEC_CHOICE_RAV1E
        AVIF_CODEC_CHOICE_SVT

    ctypedef enum avifCodecFlags:
        AVIF_CODEC_FLAG_CAN_DECODE
        AVIF_CODEC_FLAG_CAN_ENCODE

    ctypedef struct avifEncoderData:
        pass
    ctypedef struct avifCodecSpecificOptions:
        pass

    ctypedef struct avifEncoder:
        avifCodecChoice codecChoice
        int maxThreads
        int minQuantizer
        int maxQuantizer
        int minQuantizerAlpha
        int maxQuantizerAlpha
        int tileRowsLog2
        int tileColsLog2
        int speed
        int keyframeInterval    #How many frames between automatic forced keyframes; 0 to disable (default).
        uint64_t timescale      #timescale of the media (Hz)

        #stats from the most recent write
        #avifIOStats ioStats

        #Internals used by the encoder
        avifEncoderData * data
        avifCodecSpecificOptions * csOptions

    avifEncoder * avifEncoderCreate()
    avifResult avifEncoderWrite(avifEncoder * encoder, const avifImage * image, avifRWData * output)
    void avifEncoderDestroy(avifEncoder * encoder)

    ctypedef enum avifAddImageFlags:
        AVIF_ADD_IMAGE_FLAG_NONE
        AVIF_ADD_IMAGE_FLAG_FORCE_KEYFRAME
        AVIF_ADD_IMAGE_FLAG_SINGLE

    avifResult avifEncoderAddImage(avifEncoder * encoder, const avifImage * image, uint64_t durationInTimescales, uint32_t addImageFlags)
    avifResult avifEncoderAddImageGrid(avifEncoder * encoder,
                                            uint32_t gridCols,
                                            uint32_t gridRows,
                                            const avifImage * const * cellImages,
                                            uint32_t addImageFlags)
    avifResult avifEncoderFinish(avifEncoder * encoder, avifRWData * output)

    ctypedef struct avifDecoderData:
        pass
    enum avifDecoderSource:
        AVIF_DECODER_SOURCE_AUTO
        AVIF_DECODER_SOURCE_PRIMARY_ITEM
        AVIF_DECODER_SOURCE_TRACKS
    ctypedef struct avifImageTiming:
        uint64_t timescale              # timescale of the media (Hz)
        double pts                      # presentation timestamp in seconds (ptsInTimescales / timescale)
        uint64_t ptsInTimescales        # presentation timestamp in "timescales"
        double duration                 # in seconds (durationInTimescales / timescale)
        uint64_t durationInTimescales   # duration in "timescales"
    ctypedef struct avifIOStats:
        size_t colorOBUSize
        size_t alphaOBUSize
    ctypedef struct avifIO:
        pass
    ctypedef uint32_t avifStrictFlags
    ctypedef struct avifDecoder:
        # Defaults to AVIF_CODEC_CHOICE_AUTO: Preference determined by order in availableCodecs table (avif.c)
        avifCodecChoice codecChoice
        int maxThreads      # Defaults to 1
        # avifs can have multiple sets of images in them. This specifies which to decode.
        # Set this via avifDecoderSetSource().
        avifDecoderSource requestedSource

        # All decoded image data; owned by the decoder. All information in this image is incrementally
        # added and updated as avifDecoder*() functions are called. After a successful call to
        # avifDecoderParse(), all values in decoder->image (other than the planes/rowBytes themselves)
        # will be pre-populated with all information found in the outer AVIF container, prior to any
        # AV1 decoding. If the contents of the inner AV1 payload disagree with the outer container,
        # these values may change after calls to avifDecoderRead*(),avifDecoderNextImage(), or
        # avifDecoderNthImage().
        #
        # The YUV and A contents of this image are likely owned by the decoder, so be sure to copy any
        # data inside of this image before advancing to the next image or reusing the decoder. It is
        # legal to call avifImageYUVToRGB() on this in between calls to avifDecoderNextImage(), but use
        # avifImageCopy() if you want to make a complete, permanent copy of this image's YUV content or
        # metadata.
        avifImage * image
        # Counts and timing for the current image in an image sequence. Uninteresting for single image files.
        int imageIndex      # 0-based
        int imageCount      # Always 1 for non-sequences
        avifImageTiming imageTiming
        uint64_t timescale  # timescale of the media (Hz)
        double duration     # in seconds (durationInTimescales / timescale)
        uint64_t durationInTimescales   # duration in "timescales"
        # This is true when avifDecoderParse() detects an alpha plane. Use this to find out if alpha is
        # present after a successful call to avifDecoderParse(), but prior to any call to
        # avifDecoderNextImage() or avifDecoderNthImage(), as decoder->image->alphaPlane won't exist yet.
        avifBool alphaPresent
        # Enable any of these to avoid reading and surfacing specific data to the decoded avifImage.
        # These can be useful if your avifIO implementation heavily uses AVIF_RESULT_WAITING_ON_IO for
        # streaming data, as some of these payloads are (unfortunately) packed at the end of the file,
        # which will cause avifDecoderParse() to return AVIF_RESULT_WAITING_ON_IO until it finds them.
        # If you don't actually leverage this data, it is best to ignore it here.
        avifBool ignoreExif
        avifBool ignoreXMP
        uint32_t imageSizeLimit
        #uint32_t imageCountLimit
        avifStrictFlags strictFlags

        # stats from the most recent read, possibly 0s if reading an image sequence
        avifIOStats ioStats
        # Use one of the avifDecoderSetIO*() functions to set this
        avifIO * io
        # Internals used by the decoder
        avifDecoderData * data

    avifDecoder * avifDecoderCreate()
    avifResult avifDecoderSetIOMemory(avifDecoder * decoder, const uint8_t * data, size_t size)
    avifResult avifDecoderParse(avifDecoder * decoder)
    avifResult avifDecoderNextImage(avifDecoder * decoder)
    void avifDecoderDestroy(avifDecoder * decoder)

    ctypedef enum avifResult:
        AVIF_RESULT_OK
        AVIF_RESULT_UNKNOWN_ERROR
        AVIF_RESULT_INVALID_FTYP
        AVIF_RESULT_NO_CONTENT
        AVIF_RESULT_NO_YUV_FORMAT_SELECTED
        AVIF_RESULT_REFORMAT_FAILED
        AVIF_RESULT_UNSUPPORTED_DEPTH
        AVIF_RESULT_ENCODE_COLOR_FAILED
        AVIF_RESULT_ENCODE_ALPHA_FAILED
        AVIF_RESULT_BMFF_PARSE_FAILED
        AVIF_RESULT_NO_AV1_ITEMS_FOUND
        AVIF_RESULT_DECODE_COLOR_FAILED
        AVIF_RESULT_DECODE_ALPHA_FAILED
        AVIF_RESULT_COLOR_ALPHA_SIZE_MISMATCH
        AVIF_RESULT_ISPE_SIZE_MISMATCH
        AVIF_RESULT_NO_CODEC_AVAILABLE
        AVIF_RESULT_NO_IMAGES_REMAINING
        AVIF_RESULT_INVALID_EXIF_PAYLOAD
        AVIF_RESULT_INVALID_IMAGE_GRID
        AVIF_RESULT_INVALID_CODEC_SPECIFIC_OPTION
        AVIF_RESULT_TRUNCATED_DATA
        AVIF_RESULT_IO_NOT_SET
        AVIF_RESULT_IO_ERROR
        AVIF_RESULT_WAITING_ON_IO
        AVIF_RESULT_INVALID_ARGUMENT
        AVIF_RESULT_NOT_IMPLEMENTED

AVIF_RESULT = {
    AVIF_RESULT_OK                              : "OK",
    AVIF_RESULT_UNKNOWN_ERROR                   : "UNKNOWN ERROR",
    AVIF_RESULT_INVALID_FTYP                    : "INVALID_FTYP",
    AVIF_RESULT_NO_CONTENT                      : "NO_CONTENT",
    AVIF_RESULT_NO_YUV_FORMAT_SELECTED          : "NO_YUV_FORMAT_SELECTED",
    AVIF_RESULT_REFORMAT_FAILED                 : "REFORMAT_FAILED",
    AVIF_RESULT_UNSUPPORTED_DEPTH               : "UNSUPPORTED_DEPTH",
    AVIF_RESULT_ENCODE_COLOR_FAILED             : "ENCODE_COLOR_FAILED",
    AVIF_RESULT_ENCODE_ALPHA_FAILED             : "ENCODE_ALPHA_FAILED",
    AVIF_RESULT_BMFF_PARSE_FAILED               : "BMFF_PARSE_FAILED",
    AVIF_RESULT_NO_AV1_ITEMS_FOUND              : "NO_AV1_ITEMS_FOUND",
    AVIF_RESULT_DECODE_COLOR_FAILED             : "DECODE_COLOR_FAILED",
    AVIF_RESULT_DECODE_ALPHA_FAILED             : "DECODE_ALPHA_FAILED",
    AVIF_RESULT_COLOR_ALPHA_SIZE_MISMATCH       : "COLOR_ALPHA_SIZE_MISMATCH",
    AVIF_RESULT_ISPE_SIZE_MISMATCH              : "ISPE_SIZE_MISMATCH",
    AVIF_RESULT_NO_CODEC_AVAILABLE              : "NO_CODEC_AVAILABLE",
    AVIF_RESULT_NO_IMAGES_REMAINING             : "NO_IMAGES_REMAINING",
    AVIF_RESULT_INVALID_EXIF_PAYLOAD            : "INVALID_EXIF_PAYLOAD",
    AVIF_RESULT_INVALID_IMAGE_GRID              : "INVALID_IMAGE_GRID",
    AVIF_RESULT_INVALID_CODEC_SPECIFIC_OPTION   : "INVALID_CODEC_SPECIFIC_OPTION",
    AVIF_RESULT_TRUNCATED_DATA                  : "TRUNCATED_DATA",
    AVIF_RESULT_IO_NOT_SET                      : "IO_NOT_SET",
    AVIF_RESULT_IO_ERROR                        : "IO_ERROR",
    AVIF_RESULT_WAITING_ON_IO                   : "WAITING_ON_IO",
    AVIF_RESULT_INVALID_ARGUMENT                : "INVALID_ARGUMENT",
    AVIF_RESULT_NOT_IMPLEMENTED                 : "NOT_IMPLEMENTED",
    }



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import math
from collections import deque
from time import monotonic
from typing import Any, Dict, Tuple
from collections.abc import Sequence

from xpra.log import Logger
log = Logger("encoder", "vpx")

from xpra.codecs.constants import VideoSpec, get_subsampling_divs
from xpra.codecs.image import ImageWrapper
from xpra.os_util import WIN32, OSX, POSIX
from xpra.util.env import envint, envbool
from xpra.util.objects import AtomicInteger, typedict

from xpra.codecs.vpx.vpx cimport (
    vpx_img_fmt_t, vpx_codec_iface_t,
    vpx_codec_iter_t, vpx_codec_flags_t, vpx_codec_err_t,
    vpx_codec_ctx_t,
    vpx_codec_error, vpx_codec_destroy,
    vpx_codec_version_str, vpx_codec_build_config,
    VPX_IMG_FMT_I420, VPX_IMG_FMT_I444, VPX_IMG_FMT_I44416,
    vpx_image_t,
    VPX_CS_BT_709, VPX_CR_FULL_RANGE, VPX_CR_STUDIO_RANGE,
    vpx_codec_err_to_string, vpx_codec_control_,
)
from libc.stdint cimport uint8_t, int64_t
from libc.stdlib cimport free, malloc
from libc.string cimport memset


SAVE_TO_FILE = os.environ.get("XPRA_SAVE_TO_FILE")

cdef int default_nthreads = max(1, int(math.sqrt(os.cpu_count()+1)))
cdef int VPX_THREADS = envint("XPRA_VPX_THREADS", default_nthreads)

cdef inline int roundup(int n, int m):
    return (n + m - 1) & ~(m - 1)

cdef int ENABLE_VP9_TILING = envbool("XPRA_VP9_TILING", False)


cdef inline int MIN(int a, int b):
    if a<=b:
        return a
    return b

cdef inline int MAX(int a, int b):
    if a>=b:
        return a
    return b


DEF USAGE_STREAM_FROM_SERVER    = 0x0
DEF USAGE_LOCAL_FILE_PLAYBACK   = 0x1
DEF USAGE_CONSTRAINED_QUALITY   = 0x2
DEF USAGE_CONSTANT_QUALITY      = 0x3

DEF VPX_CODEC_USE_HIGHBITDEPTH = 0x40000


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "vpx/vp8cx.h":
    const vpx_codec_iface_t *vpx_codec_vp8_cx()
    const vpx_codec_iface_t *vpx_codec_vp9_cx()

cdef extern from "vpx/vpx_encoder.h":
    int VPX_ENCODER_ABI_VERSION
    #vpx_rc_mode
    int VPX_VBR         #Variable Bit Rate (VBR) mode
    int VPX_CBR         #Constant Bit Rate (CBR) mode
    int VPX_CQ          #Constant Quality (CQ) mode
    #function to set number of tile columns:
    int VP9E_SET_TILE_COLUMNS
    #function to set encoder internal speed settings:
    int VP8E_SET_CPUUSED
    #function to enable/disable periodic Q boost:
    int VP9E_SET_FRAME_PERIODIC_BOOST
    int VP9E_SET_LOSSLESS
    #vpx_enc_pass:
    int VPX_RC_ONE_PASS
    int VPX_RC_FIRST_PASS
    int VPX_RC_LAST_PASS
    #vpx_kf_mode:
    int VPX_KF_FIXED
    int VPX_KF_AUTO
    int VPX_KF_DISABLED
    long VPX_EFLAG_FORCE_KF
    ctypedef struct vpx_rational_t:
        int num     #fraction numerator
        int den     #fraction denominator
    ctypedef int vpx_bit_depth_t
    ctypedef struct vpx_codec_enc_cfg_t:
        unsigned int g_usage
        unsigned int g_threads
        unsigned int g_profile
        unsigned int g_w
        unsigned int g_h
        vpx_bit_depth_t g_bit_depth
        vpx_rational_t g_timebase
        unsigned int g_error_resilient
        unsigned int g_pass
        unsigned int g_lag_in_frames
        unsigned int rc_dropframe_thresh
        unsigned int rc_resize_allowed
        unsigned int rc_resize_up_thresh
        unsigned int rc_resize_down_thresh
        int rc_end_usage
        #struct vpx_fixed_buf rc_twopass_stats_in
        unsigned int rc_target_bitrate
        unsigned int rc_min_quantizer
        unsigned int rc_max_quantizer
        unsigned int rc_undershoot_pct
        unsigned int rc_overshoot_pct
        unsigned int rc_buf_sz
        unsigned int rc_buf_initial_sz
        unsigned int rc_buf_optimal_sz
        #we don't use 2pass:
        #unsigned int rc_2pass_vbr_bias_pct
        #unsigned int rc_2pass_vbr_minsection_pct
        #unsigned int rc_2pass_vbr_maxsection_pct
        unsigned int kf_mode
        unsigned int kf_min_dist
        unsigned int kf_max_dist
        unsigned int ts_number_layers
        unsigned int[5] ts_target_bitrate
        unsigned int[5] ts_rate_decimator
        unsigned int ts_periodicity
        unsigned int[16] ts_layer_id
    ctypedef int64_t vpx_codec_pts_t
    ctypedef long vpx_enc_frame_flags_t

    cdef int VPX_CODEC_CX_FRAME_PKT
    cdef int VPX_CODEC_STATS_PKT
    cdef int VPX_CODEC_PSNR_PKT
    cdef int VPX_CODEC_CUSTOM_PKT

    ctypedef struct frame:
        void    *buf
        size_t  sz

    ctypedef struct data:
        frame frame

    ctypedef struct vpx_codec_cx_pkt_t:
        int kind
        data data

    cdef int VPX_DL_REALTIME
    cdef int VPX_DL_GOOD_QUALITY
    cdef int VPX_DL_BEST_QUALITY

    vpx_codec_err_t vpx_codec_enc_config_default(vpx_codec_iface_t *iface,
                              vpx_codec_enc_cfg_t *cfg, unsigned int usage)
    vpx_codec_err_t vpx_codec_enc_init_ver(vpx_codec_ctx_t *ctx, vpx_codec_iface_t *iface,
                                       vpx_codec_enc_cfg_t  *cfg, vpx_codec_flags_t flags, int abi_version)

    vpx_codec_err_t vpx_codec_encode(vpx_codec_ctx_t *ctx, const vpx_image_t *img,
                              vpx_codec_pts_t pts, unsigned long duration,
                              vpx_enc_frame_flags_t flags, unsigned long deadline) nogil

    const vpx_codec_cx_pkt_t *vpx_codec_get_cx_data(vpx_codec_ctx_t *ctx, vpx_codec_iter_t *iter) nogil
    vpx_codec_err_t vpx_codec_enc_config_set(vpx_codec_ctx_t *ctx, const vpx_codec_enc_cfg_t *cfg)

PACKET_KIND: Dict[int, str] = {
    VPX_CODEC_CX_FRAME_PKT   : "CX_FRAME_PKT",
    VPX_CODEC_STATS_PKT      : "STATS_PKT",
    VPX_CODEC_PSNR_PKT       : "PSNR_PKT",
    VPX_CODEC_CUSTOM_PKT     : "CUSTOM_PKT",
}

COLORSPACES: Dict[str, Sequence[str]] = {
    "vp8": ("YUV420P", ),
}
if VPX_ENCODER_ABI_VERSION>=23:
    COLORSPACES["vp9"] = ("YUV420P", "YUV444P", "YUV444P10")
CODECS = tuple(COLORSPACES.keys())

#as of libvpx 1.8:
DEF VP9_RANGE = 4


def get_abi_version() -> int:
    return VPX_ENCODER_ABI_VERSION


def get_version() -> Sequence[int]:
    b = vpx_codec_version_str()
    vstr = b.decode("latin1").lstrip("v")
    log("vpx_codec_version_str()=%s", vstr)
    vparts : List[int] = []
    try:
        for x in vstr.split("."):
            vparts.append(int(x))
    except Exception:
        pass
    return tuple(vparts)


def get_type() -> str:
    return "vpx"


def get_encodings() -> Sequence[str]:
    return CODECS


generation = AtomicInteger()


def get_info() -> Dict[str,Any]:
    global CODECS, MAX_SIZE
    b = vpx_codec_build_config()
    info = {
        "version"       : get_version(),
        "encodings"     : CODECS,
        "abi_version"   : get_abi_version(),
        "counter"       : generation.get(),
        "build_config"  : b.decode("latin1"),
    }
    for e, maxsize in MAX_SIZE.items():
        info["%s.max-size" % e] = maxsize
    for k,v in COLORSPACES.items():
        info["%s.colorspaces" % k] = v
    return info


cdef inline const vpx_codec_iface_t  *make_codec_cx(encoding):
    if encoding=="vp8":
        return vpx_codec_vp8_cx()
    if encoding=="vp9":
        return vpx_codec_vp9_cx()
    raise ValueError(f"unsupported encoding {encoding!r}")


#educated guess:
MAX_SIZE: Dict[str, Tuple[int, int]] = {
    "vp8"   : (8192, 4096),
    "vp9"   : (8192, 4096),
}
#no idea why, but this is the default on win32:
if WIN32:
    MAX_SIZE["vp9"] = (4096, 4096)


def get_specs() -> Sequence[VideoSpec]:
    specs: Sequence[VideoSpec] = []

    for encoding in get_encodings():
        max_w, max_h = MAX_SIZE[encoding]
        if encoding=="vp8":
            has_lossless_mode = False
            speed = 50
            quality = 50
        else:
            has_lossless_mode = in_cs.startswith("YUV444P")
            speed = 40
            quality = 50 + 50*int(has_lossless_mode)

        for in_cs in COLORSPACES[encoding]:
            out_cs = in_cs
            if in_cs == "YUV444P10":
                out_cs = "r210"

            specs.append(
                VideoSpec(
                    encoding=encoding, input_colorspace=in_cs, output_colorspaces=(out_cs, ),
                    has_lossless_mode=has_lossless_mode,
                    codec_class=Encoder, codec_type=get_type(),
                    quality=quality, speed=speed,
                    size_efficiency=60,
                    setup_cost=20, max_w=max_w, max_h=max_h,
                )
            )
    return specs


cdef inline vpx_img_fmt_t get_vpx_colorspace(colorspace: str) except -1:
    if colorspace == "YUV420P":
        return VPX_IMG_FMT_I420
    if colorspace == "YUV444P":
        return VPX_IMG_FMT_I444
    if colorspace == "YUV444P10":
        return VPX_IMG_FMT_I44416
    raise ValueError(f"invalid colorspace {colorspace!r}")


def get_error_string(int err) -> str:
    estr = vpx_codec_err_to_string(<vpx_codec_err_t> err)[:]
    if not estr:
        return str(err)
    try:
        return estr.decode("latin1")
    except UnicodeDecodeError:
        return str(estr)


cdef class Encoder:
    cdef unsigned long frames
    cdef vpx_codec_ctx_t *context
    cdef vpx_codec_enc_cfg_t cfg
    cdef vpx_img_fmt_t pixfmt
    cdef unsigned int width
    cdef unsigned int height
    cdef unsigned int max_threads
    cdef unsigned int generation
    cdef unsigned long bandwidth_limit
    cdef double initial_bitrate_per_pixel
    cdef object encoding
    cdef object src_format
    cdef int speed
    cdef int quality
    cdef int lossless
    cdef object last_frame_times
    cdef object file

    cdef object __weakref__

    def init_context(self, encoding: str,
                     unsigned int width, unsigned int height,
                     src_format: str,
                     options: typedict) -> None:
        log("vpx init_context%s", (encoding, width, height, src_format, options))
        assert encoding in CODECS, "invalid encoding: %s" % encoding
        assert options.get("scaled-width", width)==width, "vpx encoder does not handle scaling"
        assert options.get("scaled-height", height)==height, "vpx encoder does not handle scaling"
        assert encoding in get_encodings()
        assert src_format in COLORSPACES.get(encoding, ()), f"invalid source format {src_format!r} for {encoding}"

        self.src_format = src_format

        cdef const vpx_codec_iface_t *codec_iface = make_codec_cx(encoding)
        cdef vpx_codec_flags_t flags = 0
        self.encoding = encoding
        self.width = width
        self.height = height
        self.quality = options.intget("quality", 50)
        self.speed = options.intget("speed", 50)
        self.bandwidth_limit = options.intget("bandwidth-limit", 0)
        self.lossless = 0
        self.frames = 0
        self.last_frame_times = deque(maxlen=200)
        self.pixfmt = get_vpx_colorspace(self.src_format)
        try:
            #no point having too many threads if the height is small, also avoids a warning:
            self.max_threads = max(0, min(int(options.intget("threads", VPX_THREADS)), roundup(height, 64)//64, 32))
        except Exception as e:
            log.error("Error parsing number of threads: %s", e)
            self.max_threads = 2

        if vpx_codec_enc_config_default(codec_iface, &self.cfg, 0)!=0:
            raise RuntimeError("failed to create vpx encoder config")
        log("%s codec defaults:", self.encoding)
        self.log_cfg()
        self.initial_bitrate_per_pixel = self.cfg.rc_target_bitrate / self.cfg.g_w / self.cfg.g_h
        log("initial_bitrate_per_pixel(%i, %i, %i)=%.3f", self.cfg.g_w, self.cfg.g_h, self.cfg.rc_target_bitrate, self.initial_bitrate_per_pixel)

        self.update_cfg()
        self.cfg.g_usage = USAGE_STREAM_FROM_SERVER
        if self.src_format=="YUV444P":
            self.cfg.g_profile = 1
            self.cfg.g_bit_depth = 8
        elif self.src_format=="YUV444P10":
            self.cfg.g_profile = 3
            self.cfg.g_bit_depth = 10
            flags |= VPX_CODEC_USE_HIGHBITDEPTH
        else:
            self.cfg.g_profile = 0
        self.cfg.g_w = width
        self.cfg.g_h = height
        cdef vpx_rational_t timebase
        timebase.num = 1
        timebase.den = 1000
        self.cfg.g_timebase = timebase
        self.cfg.g_error_resilient = 0              #we currently use TCP, guaranteed delivery
        self.cfg.g_pass = VPX_RC_ONE_PASS
        self.cfg.g_lag_in_frames = 0                #always give us compressed output for each frame without delay
        self.cfg.rc_resize_allowed = 0
        self.cfg.rc_end_usage = VPX_VBR
        #we choose when to use keyframes (never):
        self.cfg.kf_mode = VPX_KF_DISABLED
        self.cfg.kf_min_dist = 999999
        self.cfg.kf_max_dist = 999999

        self.context = <vpx_codec_ctx_t *> malloc(sizeof(vpx_codec_ctx_t))
        if self.context==NULL:
            raise RuntimeError("failed to allocate memory for vpx encoder context")
        memset(self.context, 0, sizeof(vpx_codec_ctx_t))

        log("our configuration:")
        self.log_cfg()
        cdef int ret = vpx_codec_enc_init_ver(self.context, codec_iface, &self.cfg, flags, VPX_ENCODER_ABI_VERSION)
        if ret!=0:
            free(self.context)
            self.context = NULL
            log("vpx_codec_enc_init_ver() returned %s", get_error_string(ret))
            raise RuntimeError("failed to instantiate %s encoder with ABI version %s: %s" % (encoding, VPX_ENCODER_ABI_VERSION, self.codec_error_str()))
        log("vpx_codec_enc_init_ver for %s succeeded", encoding)
        if encoding=="vp9" and ENABLE_VP9_TILING and width>=256:
            tile_columns = 0
            if width>=256:
                tile_columns = 1
            elif width>=512:
                tile_columns = 2
            elif width>=1024:
                tile_columns = 3
            self.codec_control("tile columns", VP9E_SET_TILE_COLUMNS, tile_columns)
        if encoding=="vp9":
            #disable periodic Q boost which causes latency spikes:
            self.codec_control("periodic Q boost", VP9E_SET_FRAME_PERIODIC_BOOST, 0)
        self.do_set_encoding_speed(self.speed)
        self.do_set_encoding_quality(self.quality)
        self.generation = generation.increase()
        if SAVE_TO_FILE is not None:
            filename = SAVE_TO_FILE+f"vpx-{self.generation}.{encoding}"
            self.file = open(filename, "wb")
            log.info(f"saving {encoding} stream to {filename!r}")

    def codec_error_str(self) -> str:
        return vpx_codec_error(self.context).decode("latin1")

    def codec_control(self, info, int attr, int value) -> bool:
        cdef vpx_codec_err_t ctrl = vpx_codec_control_(self.context, attr, value)
        log("%s setting %s to %s", self.encoding, info, value)
        if ctrl!=0:
            log.warn("failed to set %s to %s: %s (%s)", info, value, get_error_string(ctrl), ctrl)
        return ctrl==0

    def log_cfg(self) -> None:
        log(" target_bitrate=%s", self.cfg.rc_target_bitrate)
        log(" min_quantizer=%s", self.cfg.rc_min_quantizer)
        log(" max_quantizer=%s", self.cfg.rc_max_quantizer)
        log(" undershoot_pct=%s", self.cfg.rc_undershoot_pct)
        log(" overshoot_pct=%s", self.cfg.rc_overshoot_pct)

    cdef void update_cfg(self):
        self.cfg.rc_undershoot_pct = 100
        self.cfg.rc_overshoot_pct = 100
        bitrate_kbps = int(self.width * self.height * self.initial_bitrate_per_pixel)
        if self.bandwidth_limit>0:
            #vpx bitrate values are in Kbps:
            bitrate_kbps = min(self.bandwidth_limit//1024, bitrate_kbps)
        MIN_BITRATE = 1024
        MAX_BITRATE = 15000
        bitrate_kbps = max(MIN_BITRATE, min(MAX_BITRATE, bitrate_kbps))
        log("update_cfg() bitrate(%i,%i,%.3f,%i)=%iKbps",
            self.width, self.height, self.initial_bitrate_per_pixel, self.bandwidth_limit, bitrate_kbps)
        self.cfg.rc_target_bitrate = bitrate_kbps
        self.cfg.g_threads = self.max_threads
        self.cfg.rc_min_quantizer = MAX(0, MIN(63, int((80-self.quality) * 0.63)))
        self.cfg.rc_max_quantizer = MAX(self.cfg.rc_min_quantizer, MIN(63, int((100-self.quality) * 0.63)))

    def is_ready(self) -> bool:
        return True

    def __repr__(self):
        return "vpx.Encoder(%s)" % self.encoding

    def get_info(self) -> Dict[str,Any]:
        info = get_info()
        info |= {
            "frames"    : int(self.frames),
            "width"     : self.width,
            "height"    : self.height,
            "speed"     : self.speed,
            "quality"   : self.quality,
            "lossless"  : bool(self.lossless),
            "generation" : self.generation,
            "encoding"  : self.encoding,
            "src_format": self.src_format,
            "max_threads": self.max_threads,
            "bandwidth-limit" : int(self.bandwidth_limit),
        }
        #calculate fps:
        cdef unsigned int f = 0
        cdef double now = monotonic()
        cdef double last_time = now
        cdef double cut_off = now-10.0
        cdef double ms_per_frame = 0
        for start,end in tuple(self.last_frame_times):
            if end>cut_off:
                f += 1
                last_time = min(last_time, end)
                ms_per_frame += (end-start)
        if f>0 and last_time<now:
            info["fps"] = int(0.5+f/(now-last_time))
            info["ms_per_frame"] = int(1000.0*ms_per_frame/f)
        info |= {
            "target_bitrate"   : self.cfg.rc_target_bitrate,
            "min-quantizer"    : self.cfg.rc_min_quantizer,
            "max-quantizer"    : self.cfg.rc_max_quantizer,
            "undershoot-pct"   : self.cfg.rc_undershoot_pct,
            "overshoot-pct"    : self.cfg.rc_overshoot_pct,
        }
        return info

    def get_encoding(self) -> str:
        return self.encoding

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def is_closed(self) -> bool:
        return self.context==NULL

    def get_type(self) -> str:
        return "vpx"

    def get_src_format(self) -> str:
        return self.src_format

    def __dealloc__(self):
        self.clean()

    def clean(self) -> None:
        if self.context!=NULL:
            vpx_codec_destroy(self.context)
            free(self.context)
            self.context = NULL
        self.frames = 0
        self.pixfmt = 0
        self.width = 0
        self.height = 0
        self.max_threads = 0
        self.encoding = ""
        self.src_format = ""
        f = self.file
        if f:
            self.file = None
            f.close()

    def compress_image(self, image: ImageWrapper, options: typedict) -> Tuple[bytes, Dict]:
        cdef uint8_t *pic_in[3]
        cdef int strides[3]
        assert self.context!=NULL
        pixels = image.get_pixels()
        istrides = image.get_rowstride()
        cdef int full_range = int(image.get_full_range())
        pf = image.get_pixel_format().replace("A", "X")
        if pf != self.src_format:
            raise ValueError(f"expected {self.src_format} but got {image.get_pixel_format()}")
        assert image.get_width()==self.width, "invalid image width %s, expected %s" % (image.get_width(), self.width)
        assert image.get_height()==self.height, "invalid image height %s, expected %s" % (image.get_height(), self.height)
        assert pixels, "failed to get pixels from %s" % image
        assert len(pixels)==3, "image pixels does not have 3 planes! (found %s)" % len(pixels)
        assert len(istrides)==3, "image strides does not have 3 values! (found %s)" % len(istrides)
        cdef unsigned int Bpp = 1 + int(self.src_format.endswith("P10"))
        divs = get_subsampling_divs(self.src_format)
        cdef unsigned long bandwidth_limit = options.intget("bandwidth-limit", self.bandwidth_limit)

        if bandwidth_limit!=self.bandwidth_limit:
            self.bandwidth_limit = bandwidth_limit
            self.update_cfg()

        cdef int speed = options.intget("speed", 50)
        if speed>=0:
            self.set_encoding_speed(speed)
        cdef int quality = options.intget("quality", 50)
        if quality>=0:
            self.set_encoding_quality(quality)

        cdef Py_buffer py_buf[3]
        for i in range(3):
            memset(&py_buf[i], 0, sizeof(Py_buffer))
        try:
            for i in range(3):
                xdiv, ydiv = divs[i]
                if PyObject_GetBuffer(pixels[i], &py_buf[i], PyBUF_ANY_CONTIGUOUS):
                    raise ValueError(f"failed to read pixel data from {type(pixels[i])}")
                assert istrides[i]>=self.width*Bpp//xdiv, "invalid stride %i for width %i" % (istrides[i], self.width)
                assert py_buf[i].len>=istrides[i]*(self.height//ydiv), "invalid buffer length %i for plane %s, at least %i needed" % (
                    py_buf[i].len, "YUV"[i], istrides[i]*(self.height//ydiv))
                pic_in[i] = <uint8_t *> py_buf[i].buf
                strides[i] = istrides[i]
            return self.do_compress_image(pic_in, strides, full_range), {
                "csc"       : self.src_format,
                "frame"     : int(self.frames),
                "full-range" : bool(full_range),
                #"quality"  : min(99+self.lossless, self.quality),
                #"speed"    : self.speed,
            }
        finally:
            for i in range(3):
                if py_buf[i].buf:
                    PyBuffer_Release(&py_buf[i])

    cdef bytes do_compress_image(self, uint8_t *pic_in[3], int strides[3], int full_range):
        #actual compression (no gil):
        cdef vpx_codec_iter_t iter = NULL
        cdef int flags = 0
        cdef vpx_codec_err_t i

        cdef vpx_image_t *image = <vpx_image_t *> malloc(sizeof(vpx_image_t))
        memset(image, 0, sizeof(vpx_image_t))
        image.w = self.width
        image.h = self.height
        image.fmt = self.pixfmt
        image.cs = VPX_CS_BT_709
        image.range = VPX_CR_FULL_RANGE if full_range else VPX_CR_STUDIO_RANGE
        for i in range(3):
            image.planes[i] = pic_in[i]
            image.stride[i] = strides[i]
        image.planes[3] = NULL
        image.stride[3] = 0
        image.d_w = self.width
        image.d_h = self.height
        #this is the chroma shift for YUV420P:
        #both X and Y are downscaled by 2^1
        if self.src_format=="YUV420P":
            image.x_chroma_shift = 1
            image.y_chroma_shift = 1
        elif self.src_format.startswith("YUV444P"):
            image.x_chroma_shift = 0
            image.y_chroma_shift = 0
        else:
            raise ValueError(f"invalid colorspace {self.src_format!r}")

        image.bps = 0
        if self.frames==0:
            flags |= VPX_EFLAG_FORCE_KF
        #deadline based on speed (also affects quality...)
        cdef unsigned long deadline
        if self.speed>=90 or self.encoding=="vp9":
            deadline = VPX_DL_REALTIME
            deadline_str = "REALTIME"
        elif self.speed<10 or self.quality>=90:
            deadline = VPX_DL_BEST_QUALITY
            deadline_str = "BEST_QUALITY"
        else:
            deadline = MAX(2, VPX_DL_GOOD_QUALITY * (90-self.speed) // 100)
            #cap the deadline at 250ms, which is already plenty
            deadline = MIN(250*1000, deadline)
            deadline_str = "%8.3fms" % deadline
        cdef double start = monotonic()
        with nogil:
            ret = vpx_codec_encode(self.context, image, self.frames, 1, flags, deadline)
        if ret!=0:
            free(image)
            log.error("Error: %s encoder failure", self.encoding)
            log.error(" %r (%i)", get_error_string(ret), ret)
            return b""
        cdef double end = monotonic()
        log("vpx_codec_encode for %s took %ims (deadline=%16s for speed=%s, quality=%s)", self.encoding, 1000.0*(end-start), deadline_str, self.speed, self.quality)
        cdef const vpx_codec_cx_pkt_t *pkt
        with nogil:
            pkt = vpx_codec_get_cx_data(self.context, &iter)
        if pkt==NULL or pkt.kind!=VPX_CODEC_CX_FRAME_PKT:
            free(image)
            log.error("%s invalid packet type: %s", self.encoding, PACKET_KIND.get(pkt.kind, pkt.kind))
            return None
        self.frames += 1
        #we copy the compressed data here, we could manage the buffer instead
        #using vpx_codec_set_cx_data_buf every time with a wrapper for freeing it,
        #but since this is compressed data, no big deal
        cdef size_t size = pkt.data.frame.sz
        img = (<char*> pkt.data.frame.buf)[:size]
        free(image)
        log("vpx returning %s data: %s bytes", self.encoding, size)
        end = monotonic()
        self.last_frame_times.append((start, end))
        if self.file and size>0:
            self.file.write(img)
            self.file.flush()
        return img

    def set_encoding_speed(self, int pct) -> None:
        if self.speed==pct:
            return
        self.speed = pct
        self.do_set_encoding_speed(pct)

    cdef void do_set_encoding_speed(self, int speed):
        #Valid range for VP8: -16..16
        #Valid range for VP9: -8..8
        #But we only use positive values, negative values are just too slow
        cdef int minv = 4
        cdef int vrange = 12
        if self.encoding=="vp9":
            minv = 5
            vrange = VP9_RANGE
        #note: we don't use the full range since the percentages are mapped to -20% to +120%
        cdef int value = (speed-20)*3*vrange//200
        value = minv + MIN(vrange, MAX(0, value))
        self.codec_control("cpu speed", VP8E_SET_CPUUSED, value)

    def set_encoding_quality(self, int pct) -> None:
        if self.quality==pct:
            return
        self.quality = pct
        self.do_set_encoding_quality(pct)

    cdef void do_set_encoding_quality(self, int pct):
        self.update_cfg()
        cdef int lossless = 0
        if self.encoding=="vp9":
            if self.codec_control("lossless", VP9E_SET_LOSSLESS, pct==100):
                lossless = 1
        self.lossless = lossless
        cdef vpx_codec_err_t ret = vpx_codec_enc_config_set(self.context, &self.cfg)
        assert ret==0, "failed to updated encoder configuration, vpx_codec_enc_config_set returned %s" % ret


def selftest(full=False) -> None:
    global CODECS, SAVE_TO_FILE
    from xpra.codecs.checks import testencoder, get_encoder_max_size
    from xpra.codecs.vpx import encoder
    temp = SAVE_TO_FILE
    try:
        SAVE_TO_FILE = None
        CODECS = testencoder(encoder, full)
        #this is expensive, so don't run it unless "full" is set:
        if full and POSIX and not OSX:
            #but first, try to figure out if we have enough memory to do this
            try:
                import psutil
                freemem_MB = psutil.virtual_memory().available//1024//1024
                if freemem_MB<=4096:
                    log.info("system has only %iMB of memory available, skipping vpx max-size tests", freemem_MB)
                    full = False
                else:
                    log.info("system has %.1fGB of memory available, running full tests", freemem_MB/1024.0)
            except Exception as e:
                log.info("failed to detect free memory: %s", e)
                log.info("skipping vpx max-size tests")
                full = False
        if full:
            global MAX_SIZE
            for encoding in get_encodings():
                maxw, maxh = get_encoder_max_size(encoder, encoding, limit_w=8192, limit_h=8192)
                dmaxw, dmaxh = MAX_SIZE[encoding]
                assert maxw>=dmaxw and maxh>=dmaxh, "%s is limited to %ix%i for %s and not %ix%i" % (encoder, maxw, maxh, encoding, dmaxw, dmaxh)
                MAX_SIZE[encoding] = maxw, maxh
            log("%s max dimensions: %s", encoder, MAX_SIZE)
    finally:
        SAVE_TO_FILE = temp
